#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

mpl_version = [i for i in mpl.__version__.split('.')]
if int(mpl_version[0]) == 1:
    if int(mpl_version[1]) >= 5:
        default_cmap = 'inferno'
    else:
        default_cmap = 'afmhot'
else:
    default_cmap = 'hot'
colormap = getattr(plt.cm, default_cmap)
lcolor = colormap(95)

# Settings for plotting
font_dict = {'family': 'sans-serif',
             'style': 'normal',
             'weight': 'normal',
             'sans-serif': 'Helvetica',
             'size': 9}

fig_dict = {'figsize': (3.25, 3.25),
            'dpi': 140}

axes_dict = {'linewidth': 0.75,
             'labelpad': 8}

lines_dict = {'linewidth': 0.75,
              'markersize': 3}

mpl.rc('font', **font_dict)
mpl.rc('figure', **fig_dict)
mpl.rc('axes', **axes_dict)
mpl.rc('lines', **lines_dict)
mpl.rc('text', usetex=True)

latexpre = [r'\usepackage{siunitx}',
            r'\sisetup{detect-all}',
            r'\usepackage{helvet}',
            r'\usepackage[EULERGREEK]{sansmath}',
            r'\sansmath']
mpl.rcParams['text.latex.preamble'] = latexpre

DEFAULT_UNIT = 'nm^-1'
LABELS_DICT = {
    'raw': ['y (pixels)', 'x (pixels)'],
    'angular': ['2theta_f (deg)', 'alpha_f (deg)'],
    'rsm': ['q_xy (nm^-1)', 'q_z (nm^-1)'],
    'polar': ['q (nm^-1)', 'chi (deg)'],
    'q': ['q (nm^-1)', 'Intensity (a.u.)'],
    'chi': ['chi (deg)', 'Intensity (a.u.)'],
    'qz': ['q_z (nm^-1)', 'Intensity (a.u.)'],
    'qxy': ['q_xy (nm^-1)', 'Intensity (a.u.)']}


def show():
    """Wrapper for plt.show() to save having to separately import mpl"""
    plt.show()


def get_line_colors(cmap, ncolors):
    """
    Automatically generate a list of rgb colors for line colors in a 1D
    plot.

    Args:
        cmap: (string) name of an mpl color map (or cmap instance).
        ncolors: number of colors to be returned in the list.

    Returns:
        lcolors: a list of rgb tuples.
    """
    cmap = getattr(plt.cm, cmap)
    stop = 0.8 * 255
    lcolors = [cmap(int(i)) for i in np.linspace(0, stop, ncolors)]
    return lcolors


def get_axis_label(label=None):
    """

    Args:
        label:

    Returns:

    """
    if (label is None) or (len(label) is 0):
        return ''
    if '(' in label:
        quantity, unit = label.split('(')
        quantity = quantity.strip(' ')
        unit = unit.strip(')')
    else:
        quantity = label
        unit = None

    if not unit and quantity.startswith('q'):
        unit = DEFAULT_UNIT

    if '_' in quantity:
        quantity, quant_sub = quantity.split('_')
    else:
        quant_sub = None

    if unit and ('^' in unit):
        unit, unit_sup = unit.split('^')
    else:
        unit_sup = None

    if unit == 'um':
        unit = '\si{\micro\meter}'

    axis_label = r''
    if quantity.startswith('2'):
        axis_label += r'2'
        quantity = quantity.strip('2')
    if (quantity.lower() in ['time', 'distance', 'intensity']) or (
            'point' in quantity.lower()):
        quantity = quantity.replace(" ", "\ ")
        axis_label += '$\mathrm{%s}' % quantity
    elif quantity in ['alpha', 'beta', 'gamma', 'chi', 'psi', 'theta']:
        axis_label += r'$\%s' % quantity
    else:
        axis_label += r'$\textit{%s}' % quantity
    if quant_sub:
        axis_label += r'_{%s}' % quant_sub
    axis_label += r'$'
    if unit:
        if unit == 'deg':
            axis_label += ' ($^{\circ}$)'
        else:
            axis_label += ' ($\mathrm{%s}' % unit
            if unit_sup:
                axis_label += '^{%s}' % unit_sup
            axis_label += '$)'
    return axis_label


def plot_command(x, y, logx=False, logy=False, **kwargs):
    """
    Args:
        x:
        y:
        logx:
        logy:
        **kwargs:

    Returns:

    """
    if logy and not logx:
        line, = plt.semilogy(x, y, **kwargs)
    elif logx and not logy:
        line, = plt.semilogx(x, y, **kwargs)
    elif logx and logy:
        line, = plt.loglog(x, y, basex=10, **kwargs)
        # raise NotImplementedError('logx and logy not implemented')
    else:
        line, = plt.plot(x, y, **kwargs)
    return line


def plot(x, y,
         xlim=None, ylim=None,
         mode='q', xlabel=None, ylabel=None,
         logx=False, logy=False,
         legend=None, line_cmap=None,
         tight_layout=True, filename=None,
         show=True, newfig=True):
    """

    Args:
        x:
        y:
        xlim:
        ylim:
        labelx:
        labely:
        logx:
        logy:
        legend:
        line_cmap:
        tight_layout:
        filename:
        show:
        newfig:

    Returns:

    """
    if newfig:
        fig = plt.figure()

    kwargs = {}
    if y.__class__ == np.ndarray:
        if x.__class__ != np.ndarray:
            raise RuntimeError(('pygix.plotting.plot: y data is ndarray, x data'
                                ' is', x.__class__))
        kwargs['color'] = lcolor
        plot_command(x, y, logx, logy, **kwargs)
        xlim = (x[0], x[-1])
    elif y.__class__ == list:
        lines = []
        if line_cmap is None:
            line_cmap = default_cmap
        line_colors = get_line_colors(line_cmap, len(y))

        if x.__class__ == np.ndarray:
            for i, yy in enumerate(y):
                kwargs['color'] = line_colors[i]
                lines.append(plot_command(x, yy, logx, logy, **kwargs))
        elif x.__class__ == list:
            if len(x) is not len(y):
                raise RuntimeError('pygix.plotting.plot: x and y are lists but '
                                   'do not have the same length')
            for xx, yy in zip(x, y):
                lines.append(plot_command(xx, yy, logx, logy, **kwargs))
    else:
        raise RuntimeError('pygix.plotting.plot: y data must be numpy.ndarray'
                           ' or list of numpy.ndarrays')

    if xlim is not None:
        plt.xlim(xlim)
    if ylim is not None:
        plt.ylim(ylim)

    if mode:
        try:
            tmp_xlabel, tmp_ylabel = LABELS_DICT[mode]
        except KeyError:
            raise RuntimeError('Invalid mode for pygix.plotting.plot\n')
        if xlabel is None:
            xlabel = tmp_xlabel
        if ylabel is None:
            ylabel = tmp_ylabel
    plt.xlabel(get_axis_label(xlabel))
    plt.ylabel(get_axis_label(ylabel))

    # set the apsect ratio to square, problem is, if you use the
    # zoom function on the widget the data apsect remains the same
    # ax = plt.gca()
    # x0, x1 = ax.get_xlim()
    # y0, y1 = ax.get_ylim()
    # ax.set_aspect((x1-x0)/(y1-y0))

    # need to handle position with bbox_to_anchor
    if legend:
        if len(legend) != len(lines):
            raise RuntimeError('pygix.plotting.plot: not enough legend labels'
                               ' for number of lines')
        plt.legend(lines, legend)

    if tight_layout:
        plt.tight_layout()

    if filename is not None:
        plt.savefig(filename, dpi=300, bbox_inches='tight')

    if show:
        plt.show()


# def rsmplot(data, x=None, y=None, xlim=None, ylim=None,
#             xlabel=None, ylabel=None, cmap=None, colorbar=False, clim='auto',
#             newfig=True, show=True, figsize=(7, 7), tight_layout=True,
#             filename=None, **kwargs):


def implot(data, x=None, y=None, mode=None,
           xlim=None, ylim=None, xlabel=None, ylabel=None,
           cmap=None, clim='auto', colorbar=False,
           newfig=True, show=True, tight_layout=True,
           filename=None,
           **kwargs):
    """
       Parameters
    ----------
    data : ndarray
        Image data array
    x : ndarray
        x axis scale
    y : ndarray
        y axis scale
    mode : string
        Data type. Accepted values: raw, angular, rsm
    xlim : tuple
        (lower, upper) limit of x axis
    ylim : tuple
        (lower, upper) limit of y axis
    cmap : string
        Valid name of matplotlib colormap
    colorbar : bool
        Add colorbar or not
    clim : 'auto', None or tuple
        z-scaling of image data. If 'auto' will set based on
        (0.5, 99.5) percentile of intensity values; if None uses
        (data.min(), data.max()); otherwise tuple of (min, max).
    newfig : bool
        Specify whether to generate a new figure or not. This is
        useful if you are creating a figure with many subplots,
        using newfig=False will not make a new figure window.
    show : bool
        Specify whether to call plt.show() when plot has been
        created. Again, useful for subplots or batch processing.
    filename : string or None
        If a string is passed the plot will be saved with given
        filename.

    Returns
    -------
    fig : matplotlib.pyplot.figure instance
        If newfig=True, returns the figure object.
    im : matplotlib.pyplot.imshow instance
        Returns only im object if newfig=False, otherwise
        (fig, im).
    """
    if mode and (mode not in ['raw', 'angular', 'rsm', 'polar']):
        raise RuntimeError('pygix.plotting.implot: mode must be raw, angular'
                           ' or rsm')
    if mode == 'raw':
        origin = 'upper'
    else:
        origin = 'lower'

    try:
        xmin, xmax = x[0], x[-1]
    except TypeError:
        xmin, xmax = 0, data.shape[1]
    try:
        ymin, ymax = y[0], y[-1]
    except TypeError:
        ymin, ymax = 0, data.shape[0]
    extent = [xmin, xmax, ymin, ymax]

    if not cmap:
        cmap = default_cmap

    if clim is None:
        vmin, vmax = data.min(), data.max()
    elif clim == 'auto':
        vmin, vmax = np.percentile(data, (0.1, 99.5))
    else:
        vmin, vmax = clim

    if newfig:
        fig = plt.figure()

    if mode == 'polar':
        if kwargs is None:
            kwargs = {}
        kwargs['aspect'] = 0.2

    im = plt.imshow(data,
                    origin=origin,
                    vmin=vmin, vmax=vmax,
                    interpolation='nearest',
                    cmap=cmap,
                    extent=extent,
                    **kwargs)
    ax = plt.gca()

    if xlim:
        plt.xlim(xlim[0], xlim[1])
    if ylim:
        plt.ylim(ylim[0], ylim[1])

    if mode:
        try:
            tmp_xlabel, tmp_ylabel = LABELS_DICT[mode]
        except KeyError:
            raise RuntimeError('Invalid mode for pygix.plotting.implot\n')
        if xlabel is None:
            xlabel = tmp_xlabel
        if ylabel is None:
            ylabel = tmp_ylabel
    plt.xlabel(get_axis_label(xlabel))
    plt.ylabel(get_axis_label(ylabel))

    if colorbar:
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im, cax=cax)

    ax.set_axis_bgcolor(colormap(0))

    if tight_layout:
        plt.tight_layout()

    if filename is not None:
        plt.savefig(filename, dpi=300, bbox_inches='tight')

    if show:
        plt.show()

    if newfig:
        return fig, im, ax
    else:
        return im, ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Main module of pygix that users will interact with.
"""
from __future__ import absolute_import, print_function, division

__authors__ = ["Thomas Dane", "Jérôme Kieffer"]
__license__ = "GPLv3+"
__date__ = "18/03/2016"
__contact__ = "thomasgdane@gmail.com"
__license__ = "GPLv3+"
__copyright__ = "2016 ESRF - The European Synchrotron, Grenoble, France"
__status__ = "Development"
__docformat__ = "restructuredtext"

import os
import logging

logger = logging.getLogger("pygix.transform")
import types
import threading
import gc
import numpy as np
from math import pi

EPS32 = (1.0 + np.finfo(np.float32).eps)

from . import grazing_geometry

GrazingGeometry = grazing_geometry.GrazingGeometry
from . import grazing_units
from . import io
import pyFAI
import fabio

error = None

try:
    from pyFAI.ext import splitBBoxLUT
except ImportError as error:  # IGNORE:W0703
    logger.warning("Unable to import pyFAI.splitBBoxLUT for"
                   " Look-up table based azimuthal integration")
    splitBBoxLUT = None

try:
    from pyFAI.ext.fastcrc import crc32
except ImportError:
    from zlib import crc32

try:
    # Used for 1D integration
    from pyFAI.ext import splitPixel  # IGNORE:F0401
except ImportError as error:
    logger.error("Unable to import pyFAI.splitPixel"
                 " full pixel splitting: %s" % error)
    splitPixel = None

try:
    from pyFAI.ext import splitBBox  # IGNORE:F0401
except ImportError as error:
    logger.error("Unable to import pyFAI.ext.splitBBox"
                 " Bounding Box pixel splitting: %s" % error)
    splitBBox = None

try:
    from pyFAI.ext import histogram
except ImportError as error:
    logger.error("Unable to import pyFAI.ext.histogram"
                 " Cython histogram implementation: %s" % error)
    histogram = None

try:
    from pyFAI.ext import splitBBoxCSR  # IGNORE:F0401
except ImportError as error:
    logger.error("Unable to import pyFAI.ext.splitBBoxCSR"
                 " CSR based azimuthal integration: %s" % error)
    splitBBoxCSR = None

try:
    from pyFAI.ext import splitPixelFullCSR  # IGNORE:F0401
except ImportError as error:
    logger.error("Unable to import pyFAI.ext.splitPixelFullCSR"
                 " CSR based azimuthal integration: %s" % error)
    splitPixelFullCSR = None

from pyFAI.opencl import ocl

if ocl:
    try:
        from pyFAI import ocl_azim  # IGNORE:F0401
    except ImportError as error:  # IGNORE:W0703
        logger.warning("Unable to import pyFAI.ocl_azim"
                       ": %s" % error)
        ocl_azim = None
    try:
        from pyFAI import ocl_azim_csr  # IGNORE:F0401
    except ImportError as error:
        logger.error("Unable to import pyFAI.ocl_azim_csr"
                     ": %s" % error)
        ocl_azim_csr = None
    try:
        from pyFAI import ocl_azim_lut  # IGNORE:F0401
    except ImportError as error:  # IGNORE:W0703
        logger.error("Unable to import pyFAI.ocl_azim_lut for"
                     ": %s" % error)
        ocl_azim_lut = None
    try:
        from pyFAI import ocl_sort
    except ImportError as error:  # IGNORE:W0703
        logger.error("Unable to import pyFAI.ocl_sort for"
                     ": %s" % error)
        ocl_sort = None
else:
    ocl_azim = ocl_azim_csr = ocl_azim_lut = None


class Transform(GrazingGeometry):
    """
    This class contains methods for performing transformation and
    integration of images collected in the grazing-incidence geometry.
 
    The operations are applicable to the related techniques of:
    - grazing-incidence X-ray diffraction (GIXRD, GID)
    - grazing-incidence wide-angle X-ray scattering (GIWAXS)
    - grazing-incidence small-angle X-ray scattering (GISAXS)
    These techniques will collectively be referred to as grazing-
    incidence X-ray scattering (GIXS).

    The class is based on pyFAI's AzimuthalIntegrator class. 
    GrazingTransform inherets from GrazingGeometry, which itself 
    inherets from pyFAI's Geometry class. All geometry calculation
    are done in the GrazingGeometry or pyFAI Geometry class.

    The development of these classes was necessary to introduce the 
    reciprocal space transformation equations (see grazingGeometry) 
    and to handle additional variables that could not be accounted for 
    in pyFAI:
      - grazingGeometry : the orientaition of the surface plane 
        (defines orientaition and sign of q vectors relative to input 
         image)
      - alpha_i :  the incident angle
      - misalign : an angle defining a misaligned surface plane (i.e. 
        not perfectly horizontal/vertical)

    These variables are class attributes to GrazingTransform. This is
    beneficial if the geometry and incident angle are constant (or at 
    least there is a commonly used incident angle) as look-up tables
    (LUTs) describing transformation and integration can be stored,
    speeding up processing time. When the geometry, incident angle or
    misalignment angle change either the class attributes can be 
    changed using the relevant getters or their values can be passed
    directly to the transformation/integration functions (therefore 
    not altering the class attributes).

    The grazing-incidence geometry ("giGeometry") can take one of 4 
    int values (1, 2, 3 or 4) is defined as follows:
        1: Sample plane horizontal; bottom-top = +ve Qz
        2: Sample plane vertical;   left-right = +ve Qz
        3: Sample plane horizontal; bottom-top = -ve Qz
        4: Sample plane vertical;   left-right = -ve Qz

    N.B. that it is not possible to rotate images before-hand as this 
    will nullify the poni file (as well as darks, flats, splines etc).
    For more details see the main documentation.
    
    The main methods are:
        >>> I, qxy, qz = pg.transform_reciprocal(data)
                Transform raw image to angular or q coordinates
        >>> I, q, chi  = pg.transform_polar(data)
                Transform image to cake plot (chi vs q or 2theta)
        >>> I, q = gi.integrate_1d(data, npt)
                1d integration
    
    There are four wrappers for the 1d integration:
        >>> profile_sector
        >>> profile_chi
        >>> profile_op_box
        >>> profile_ip_box

    The error can be calculated and returned for each method if either
    a variance ndarray or error model is provided.

    For each method the resulting intensty data is always taken 
    directly from the source raw image (e.g. rather than performing 2d 
    integration on GIXS reciprocal space maps). This prevents mutliple 
    re-sampling, which can lead to the smearing of reflections and 
    inaccurate results.
    """

    DEFAULT_METHOD = "bbox"

    def __init__(self, dist=1, poni1=0, poni2=0, rot1=0, rot2=0, rot3=0,
                 pixel1=0, pixel2=0, splinefile=None, detector=None,
                 wavelength=None,
                 useqx=True, sample_orientation=1, incident_angle=None,
                 tilt_angle=0):
        """
        Parameters
        ----------
        dist : float
            Distance from sample to detector plane (orthogonal 
            distance, not along the beam), in meters.
        poni1 : float
            Coordinate of the point-of-normal-incidence along the 
            detector's first dimension, in meters.
        poni2 : float
            Coordinate of the point-of-normal-incidence along the 
            detector's second dimension, in meters.
        rot1 : float
            First rotation from sample reference to detector 
            reference, in radians.
        rot2 : float
            Second rotation from sample reference to detector 
            reference, in radians.
        rot3 : float
            Third rotation from sample reference to detector 
            reference, in radians.
        pixel1 : float
            Pixel size of the fist dimension of the detector, in 
            meters.
        pixel2 : float
            Pixel size of the second dimension of the detector, in 
            meters.
        splinefile : str
            File containing the geometric distortion of the detector. 
            Overrides the pixel size.
        detector : str or pyFAI.detector
            Name of the detector or Detector instance.
        wavelength : float
            X-ray wavelength in meters.
        sample_orientation : int
            Orientation of surface plane relative to image axes. Can 
            be 0, 1, 2 or 3. 
        incident_angle : float
            Incident angle (degrees).
        tilt_angle : float
            Misalignment tilt angle of surface plane (degrees).
        """
        GrazingGeometry.__init__(self, dist, poni1, poni2, rot1, rot2, rot3,
                                 pixel1, pixel2, splinefile, detector,
                                 wavelength,
                                 useqx, sample_orientation, incident_angle,
                                 tilt_angle)

        self._flatfield = None
        self._darkcurrent = None
        self._flatfield_crc = None
        self._darkcurrent_crc = None
        self.flatfiles = None
        self.darkfiles = None

        self.header = None

        # self._ocl_integrator = None
        # self._ocl_lut_integr = None
        # self._ocl_csr_integr = None
        # self._lut_integrator = None
        # self._csr_integrator = None
        # self._ocl_sem = threading.Semaphore()
        # self._lut_sem = threading.Semaphore()
        # self._csr_sem = threading.Semaphore()
        # self._ocl_csr_sem = threading.Semaphore()
        # self._ocl_lut_sem = threading.Semaphore()

        self._lut_sem = threading.Semaphore()
        self._ocl_lut_sem = threading.Semaphore()

        # Look-up tables for gi transformation
        self._ocl_lut_gi_transformer = None
        self._lut_gi_transformer = None
        self._csr_gi_transformer = None

        # # Look-up tables for gi integration
        self._ocl_lut_gi_integrator = None
        self._lut_gi_integrator = None
        self._empty = 0.0

        self._last_method = None

    def reset(self):
        """
        Reset GrazingTransform class. 
        """
        GrazingGeometry.reset(self)

        try:
            with self._lut_sem:
                self._lut_gi_transformer = None
                self._lut_gi_integrator = None
        except AttributeError:
            pass

    def make_mask(self, data, mask=None, dummy=None, delta_dummy=None,
                  mode="normal"):
        """
        Taken from:
        pyFAI.azimuthalIntegrator.AzimuthalIntegrator.mask_mask

        This method combine two masks (dynamic mask from *data &
        dummy* and *mask*) to generate a new one with the 'or' binary
        operation.  One can adjust the level, with the *dummy* and
        the *delta_dummy* parameter, when you consider the *data*
        values needs to be masked out.

        This method can work in two different *mode*:

            * "normal": False for valid pixels, True for bad pixels
            * "np": True for valid pixels, false for others

        This method tries to accomodate various types of masks (like
        valid=0 & masked=-1, ...) and guesses if an input mask needs
        to be inverted.

        Parameters
        ----------
        data : ndarray
            Input array of data.
        mask : ndarray
            Input mask (if None, self.mask is used).
        dummy : float
            Dummy value of dead pixels.
        delta_dummy : float
            Precision of dummy pixels.
        mode : str
            Mode can be "normal" or "np" (inverted) or "where" 
            applied to the mask.

        Returns
        -------
        mask : ndarray of bool
            The new mask
        """
        shape = data.shape
        if mask is None:
            mask = self.mask
        if mask is None:
            mask = np.zeros(shape, dtype=bool)
        elif (mask.min() < 0) and (
            mask.max() == 0):  # 0 is valid, <0 is invalid
            mask = (mask < 0)
        else:
            mask = mask.astype(bool)
        if mask.sum(dtype=int) > mask.size // 2:
            logger.warning("Mask likely to be inverted as more"
                           " than half pixel are masked !!!")
            np.logical_not(mask, mask)
        if mask.shape != shape:
            try:
                mask = mask[:shape[0], :shape[1]]
            except Exception as error:  # IGNORE:W0703
                logger.error("Mask provided has wrong shape:"
                             " expected: %s, got %s, error: %s" %
                             (shape, mask.shape, error))
                mask = np.zeros(shape, dtype=bool)
        if dummy is not None:
            if delta_dummy is None:
                np.logical_or(mask, (data == dummy), mask)
            else:
                np.logical_or(mask, abs(data - dummy) <= delta_dummy, mask)
        if mode == "np":
            np.logical_not(mask, mask)
        elif mode == "where":
            mask = np.where(np.logical_not(mask))
        return mask

    def gi_setup_lut(self, shape, npt, process, mask=None,
                     pos0_range=None, pos1_range=None,
                     mask_checksum=None, unit=grazing_units.Q):
        """
        Prepare a look-up-table for LUT-based methods. Used for
        giTransformImage, giIntegrate2d and giIntegrate1d.
        """
        if ("__len__" in dir(npt)) and (len(npt) == 2):
            int2d = True
        else:
            int2d = False

        pos1, pos0 = self.giarray_from_unit(shape, process, "center", unit)
        dpos1, dpos0 = self.giarray_from_unit(shape, process, "delta", unit)

        if ("__len__" in dir(pos0_range)) and (len(pos0_range) > 1):
            pos0_min = min(pos0_range)
            pos0_maxin = max(pos0_range)
            default_pos0 = False
        else:
            pos0_min = pos0.min()
            pos0_maxin = pos0.max()
            default_pos0 = True
        if ("__len__" in dir(pos1_range)) and (len(pos1_range) > 1):
            pos1_min = min(pos1_range)
            pos1_maxin = max(pos1_range)
            default_pos1 = False
        else:
            pos1_min = pos1.min()
            pos1_maxin = pos1.max()
            default_pos1 = True

        pos0Range = (pos0_min, pos0_maxin * EPS32)
        pos1Range = (pos1_min, pos1_maxin * EPS32)

        if mask is None:
            mask_checksum = None
        else:
            assert mask.shape == shape

        if int2d:
            gi_lut = splitBBoxLUT.HistoBBox2d(pos0, dpos0, pos1, dpos1,
                                              bins=npt,
                                              pos0Range=pos0Range,
                                              pos1Range=pos1Range,
                                              mask=mask,
                                              mask_checksum=mask_checksum,
                                              allow_pos0_neg=True,
                                              unit=unit)
        else:
            gi_lut = splitBBoxLUT.HistoBBox1d(pos0, dpos0, pos1, dpos1,
                                              bins=npt,
                                              pos0Range=pos0Range,
                                              pos1Range=pos1Range,
                                              mask=mask,
                                              mask_checksum=mask_checksum,
                                              allow_pos0_neg=True,
                                              unit=unit)

        return gi_lut, default_pos0, default_pos1

    csr_progress = """
    def gi_setup_CSR(self, shape, npt, process, mask=None,
                     pos0_range=None, pos1_range=None,
                     mask_checksum=None, unit=grazing_units.Q, split="bbox"):
        \"""
        Prepare a look-up-table for LUT-based methods. Used for
        giTransformImage, giIntegrate2d and giIntegrate1d.
        \"""
        if "__len__" in dir(npt) and len(npt) == 2:
            int2d = True
        else:
            int2d = False
        if split == "full":
            pos = self.giarray_from_unit(shape, process, "corner", unit)
        else:
            pos1, pos0 = self.giarray_from_unit(shape, process, "center", unit)
            if split = "no": 
                dpos1, dpos0 = (None, None)
            else:
                dpos1, dpos0 = self.giarray_from_unit(shape, process, 
                                                      "delta", unit)
            if (pos1_range is None) and (not int2d):
                
                
                
        pos1, pos0   = self.giarray_from_unit(shape, process, "center", unit)
        dpos1, dpos0 = self.giarray_from_unit(shape, process, "delta", unit)

        if ("__len__" in dir(pos0_range)) and (len(pos0_range) > 1):
            pos0_min = min(pos0_range)
            pos0_maxin = max(pos0_range)
            default_pos0 = False
        else:
            pos0_min = pos0.min()
            pos0_maxin = pos0.max()
            default_pos0 = True
        if ("__len__" in dir(pos1_range)) and (len(pos1_range) > 1):
            pos1_min = min(pos1_range)
            pos1_maxin = max(pos1_range)
            default_pos1 = False
        else:
            pos1_min = pos1.min()
            pos1_maxin = pos1.max()
            default_pos1 = True

        pos0Range = (pos0_min, pos0_maxin * EPS32)
        pos1Range = (pos1_min, pos1_maxin * EPS32)
        
        if mask is None:
            mask_checksum = None
        else:
            assert mask.shape == shape

        if int2d:
            gi_lut = splitBBoxLUT.HistoBBox2d(pos0, dpos0, pos1, dpos1,
                                              bins=npt,
                                              pos0Range=pos0Range,
                                              pos1Range=pos1Range,
                                              mask=mask,
                                              mask_checksum=mask_checksum,
                                              allow_pos0_neg=True,
                                              unit=unit)
        else:
            gi_lut = splitBBoxLUT.HistoBBox1d(pos0, dpos0, pos1, dpos1,
                                              bins=npt,
                                              pos0Range=pos0Range,
                                              pos1Range=pos1Range,
                                              mask=mask,
                                              mask_checksum=mask_checksum,
                                              allow_pos0_neg=True,
                                              unit=unit)

        return gi_lut, default_pos0, default_pos1
    """

    def giarray_from_unit(self, shape, process, typ, unit):
        """
        Generate an array of coordinates in different dimensions 
        (tthf_alpf, qxy_qz or qy_qz) for GIXS transformation.

        process = ["angular", "reciprocal", "polar",
                   "sector", "chi", "ipbox", "opbox"]
       
           | methods:              |   process name:
        ----------------------------------------------
        2D | transform_angular     |  angular
           | transform_reciprocal  |  reciprocal
           | transform_polar       |  polar
        ----------------------------------------------- 
        1D | profile_sector        |  sector
           | profile_chi           |  chi
           | profile_ip_box        |  ipbox
           | profile_op_box        |  opbox
       
        """
        if not typ in ("center", "corner", "delta"):
            raise RuntimeError(("Unkown type of array %s," % typ))

        if process in ["polar", "sector", "chi"]:
            typ = "abs_" + typ

        unit = grazing_units.to_unit(unit)
        out = GrazingGeometry.__dict__[unit[typ]](self, shape)

        if process in ["chi", "opbox"]:
            if "corner" in typ:
                temp = np.zeros((shape[0], shape[1], 4, 2))
                temp[:, :, :, 0] = out[:, :, :, 1]
                temp[:, :, :, 1] = out[:, :, :, 0]
                out = temp
            else:
                out = (out[1], out[0])
        return out

    # --------------------------------------------------------------------------
    #   Reciprocal space transformation function
    # --------------------------------------------------------------------------

    def transform_angular(self, data, npt=None,
                          ip_range=None, op_range=None,
                          filename=None, correctSolidAngle=False,
                          variance=None, error_model=None,
                          mask=None, dummy=None, delta_dummy=None,
                          polarization_factor=None, dark=None, flat=None,
                          method='splitpix', unit='deg',
                          safe=True, normalization_factor=None, all=False):
        """ 
        Wrapper for transform_image to project input array into
        angular coordinates I(2th_f,a_f).
        """
        unit = '2theta_%s' % unit
        out = self.transform_image(
            data,
            process='angular',
            npt=npt,
            x_range=ip_range,
            y_range=op_range,
            filename=filename,
            correctSolidAngle=correctSolidAngle,
            variance=variance,
            error_model=error_model,
            mask=mask,
            dummy=dummy,
            delta_dummy=delta_dummy,
            polarization_factor=polarization_factor,
            dark=dark,
            flat=flat,
            method=method,
            unit=unit,
            safe=safe,
            normalization_factor=normalization_factor,
            all=all
        )
        return out

    def transform_reciprocal(self, data, npt=None,
                             ip_range=None, op_range=None,
                             filename=None, correctSolidAngle=False,
                             variance=None, error_model=None,
                             mask=None, dummy=None, delta_dummy=None,
                             polarization_factor=None, dark=None, flat=None,
                             method='splitpix', unit='nm',
                             safe=True, normalization_factor=None, all=False):

        """         
        Wrapper for transform_image to project input array into
        reciprocal space coordinates I(q_xy,q_z).
        """
        unit = 'q_%s^-1' % unit
        out = self.transform_image(
            data,
            process='reciprocal',
            npt=npt,
            x_range=ip_range,
            y_range=op_range,
            filename=filename,
            correctSolidAngle=correctSolidAngle,
            variance=variance,
            error_model=error_model,
            mask=mask,
            dummy=dummy,
            delta_dummy=delta_dummy,
            polarization_factor=polarization_factor,
            dark=dark,
            flat=flat,
            method=method,
            unit=unit,
            safe=safe,
            normalization_factor=normalization_factor,
            all=all
        )
        return out

    def transform_polar(self, data, npt=(1000, 200),
                        q_range=None, chi_range=(-100, 100),
                        filename=None, correctSolidAngle=False,
                        variance=None, error_model=None,
                        mask=None, dummy=None, delta_dummy=None,
                        polarization_factor=None, dark=None, flat=None,
                        method='splitpix', unit='nm',
                        safe=True, normalization_factor=None, all=False):
        """ 
        Wrapper for transform_image to project input array into
        polar coordinates I(q,chi).
        """
        unit = 'q_%s^-1' % unit
        out = self.transform_image(
            data, process='polar',
            npt=npt,
            x_range=q_range,
            y_range=chi_range,
            filename=filename,
            correctSolidAngle=correctSolidAngle,
            variance=variance,
            error_model=error_model,
            polarization_factor=polarization_factor,
            dark=dark,
            flat=flat,
            method=method,
            unit=unit,
            safe=safe,
            normalization_factor=normalization_factor,
            all=all
        )
        return out

    def transform_image(self, data, process='transform', npt=None,
                        x_range=None, y_range=None,
                        filename=None, correctSolidAngle=False,
                        variance=None, error_model=None,
                        mask=None, dummy=None, delta_dummy=None,
                        polarization_factor=None, dark=None, flat=None,
                        method='splitpix', unit=grazing_units.Q,
                        safe=True, normalization_factor=None, all=False):
        """
        Project a raw GIXS pattern into reciprocal space qxy(nm^-1) vs 
        qz(nm^-1) by default. Based on pyFAI integrate 2d. Multi 
        algorithm implementation (tries to be bullet proof).

        Parameters
        ----------
        data : ndarray
            2D array from detector (raw image).
        filename : str
            Output filename in 2/3 column ascii format.
        correctSolidAngle : bool
            Correct for solid angle of each pixel if True.
        variance : ndarray
            Array containing the variance of the data. If not available, 
            no error propagation is done.
        error_model : str
            When variance is unknown, an error model can be given: 
            "poisson" (variance = I), "azimuthal" (variance = 
            (I-<I>)^2).
        x_range : (float, float), optional
            The lower and upper unit of the in-plane unit. If not 
            provided, range is simply (data.min(), data.max()). Values 
            outside the range are ignored.
        y_range : (float, float), optional
            The lower and upper range of the out-of-plane unit. If not 
            provided, range is simply (data.min(), data.max()). Values 
            outside the range are ignored.
        mask : ndarray
            Masked pixel array (same size as image) with 1 for masked 
            pixels and 0 for valid pixels.
        dummy : float
            Value for dead/masked pixels.
        delta_dummy : float
            Precision for dummy value
        polarization_factor : float
            Polarization factor between -1 and +1. 0 for no correction.
        dark : ndarray
            Dark current image.
        flat : ndarray
            Flat field image.
        method : str
            Integration method. Can be "np", "cython", "bbox",
            "splitpix", "lut" or "lut_ocl" (if you want to go on GPU).
        unit : str
            Grazing-incidence units. Can be "2th_af_deg", "2th_af_rad", 
            "qxy_qz_nm^-1", "qxy_qz_A^-1", "qxy_qz_nm^-1" or 
            "qxy_qz_A^-1". For GISAXS qy vs qz is typically preferred;
            for GIWAXS, qxy vs qz.
            (TTH_AF_DEG, TTH_AF_RAD, QY_QZ_NM, QY_QZ_A, QXY_QZ_NM, 
            QXY_QZ_A).
        safe : bool
            Do some extra check to ensure LUT is still valid. False is
            faster.
        normalization_factor : float
            Value of a normalization monitor.

        Returns
        -------
        I, bins_x, bins_y : 3-tuple of ndarrays
            Regrouped intensity, in-plane bins, out-of-plane bins.

        or
        
        I, bins_x, bins_y, sigma : 4-tuple of ndarrays
            Regrouped intensity, in-plane bins, out-of-plane bins and 
            error.
        """
        method = method.lower()
        if (method != self._last_method) and (self._last_method is not None):
            self._transformedmask = None

        unit = grazing_units.to_unit(unit)
        pos_scale = unit.scale
        if mask is None:
            mask = self.mask
        shape = data.shape

        if (npt is None) and ('polar' not in process):
            if self._sample_orientation in [1, 3]:
                npt = (shape[1], shape[0])
            else:
                npt = (shape[0], shape[1])

        if x_range:
            x_range = tuple([i / pos_scale for i in x_range])
        if 'polar' in process:
            if y_range is None:
                y_range = (-90, 90)
            if npt is None:
                npt = (1000, (abs(y_range[0]) + abs(y_range[1])))
            y_range = tuple([np.deg2rad(i) + pi for i in y_range])
        else:
            if y_range:
                y_range = tuple([i / pos_scale for i in y_range])

        if variance is not None:
            assert variance.size == data.size
        elif error_model:
            error_model = error_model.lower()
            if error_model == "poisson":
                variance = np.ascontiguousarray(data, np.float32)

        if correctSolidAngle:
            solidangle = self.solidAngleArray(shape, correctSolidAngle)
        else:
            solidangle = None

        if polarization_factor is None:
            polarization = None
        else:
            polarization = self.polarization(shape, polarization_factor)

        if dark is None:
            dark = self.darkcurrent

        if flat is None:
            flat = self.flatfield

        I = None
        sigma = None
        sum = None
        count = None

        if (I is None) and ("lut" in method):
            logger.debug("in lut")
            mask_crc = None
            with self._lut_sem:
                reset = None
                if self._lut_gi_transformer is None:
                    reset = "init"
                    if mask is None:
                        mask = self.detector.mask
                        mask_crc = self.detector._mask_crc
                    else:
                        mask_crc = crc32(mask)
                if (not reset) and safe:
                    if mask is None:
                        mask = self.detector.mask
                        mask_crc = self.detector._mask_crc
                    else:
                        mask_crc = crc32(mask)
                    if self._lut_gi_transformer.unit != unit:
                        reset = "unit changed"
                    if self._lut_gi_transformer.bins != npt:
                        reset = "number of points changed"
                    if self._lut_gi_transformer.size != data.size:
                        reset = "input image size changed"
                    if (mask is not None) and (
                    not self._lut_gi_transformer.check_mask):
                        reset = "mask but LUT was without mask"
                    elif (mask is None) and self._lut_gi_transformer.check_mask:
                        reset = "no mask but LUT has mask"
                    elif (mask is not None) and (
                        self._lut_gi_transformer.mask_checksum != mask_crc):
                        reset = "mask changed"
                    if (x_range is None) and (
                    not self._lut_gi_transformer.default_ipl):
                        reset = "x_range was defined in LUT"
                    elif (x_range is not None) and \
                            (self._lut_gi_transformer.pos0Range != (
                            min(x_range), max(x_range) * EPS32)):
                        reset = ("x_range is defined"
                                 " but not the same as in LUT")
                    if (
                        y_range is None) and not self._lut_gi_transformer.default_opl:
                        reset = "y_range not defined and LUT had y_range defined"
                    elif (y_range is not None) and \
                            (self._lut_gi_transformer.pos1Range != (
                            min(y_range), max(y_range) * EPS32)):
                        reset = "y_range requested and LUT's y_range don't match"
                error = False
                if reset:
                    logger.info(
                        "pygix.Transform.transform_image: Resetting transformer because %s" % reset)
                    try:
                        lut_gi_transformer, def_ip, def_op = \
                            self.gi_setup_lut(shape, npt,
                                              "transform", mask,
                                              x_range,
                                              y_range,
                                              mask_checksum=mask_crc,
                                              unit=unit)

                        self._lut_gi_transformer = lut_gi_transformer
                        self._lut_gi_transformer.default_ipl = def_ip
                        self._lut_gi_transformer.default_opl = def_op
                        error = False
                    except MemoryError:  # LUT method is hungry...
                        logger.warning("MemoryError: falling back on"
                                       " forward implementation")
                        self._ocl_lut_gi_transformer = None
                        gc.collect()
                        method = "splitbbox"
                        error = True
                if not error:
                    if ("ocl" in method) and ocl_azim_lut:
                        with self._ocl_lut_sem:
                            if "," in method:
                                c = method.index(",")
                                platformid = int(method[c - 1])
                                deviceid = int(method[c + 1])
                                devicetype = "all"
                            elif "gpu" in method:
                                platformid = None
                                deviceid = None
                                devicetype = "gpu"
                            elif "cpu" in method:
                                platformid = None
                                deviceid = None
                                devicetype = "cpu"
                            else:
                                platformid = None
                                deviceid = None
                                devicetype = "all"
                            if (self._ocl_lut_gi_transformer is None) or \
                                    (self._ocl_lut_gi_transformer.on_device[
                                         "lut"] \
                                             != self._lut_gi_transformer.lut_checksum):
                                self._ocl_lut_gi_transformer = \
                                    ocl_azim_lut.OCL_LUT_Integrator(
                                        self._lut_gi_transformer.lut,
                                        self._lut_gi_transformer.size,
                                        devicetype=devicetype,
                                        platformid=platformid,
                                        deviceid=deviceid,
                                        checksum=self._lut_gi_transformer.lut_checksum)

                            if not error:
                                I, sum, count = self._ocl_lut_gi_transformer.integrate(
                                    data, dark=dark, flat=flat,
                                    solidAngle=solidangle,
                                    solidAngle_checksum=self._dssa_crc,
                                    dummy=dummy, delta_dummy=delta_dummy,
                                    polarization=polarization,
                                    polarization_checksum=self._polarization_crc,
                                    safe=safe)
                            I.shape = npt
                            I = I.T

                            bins_x = self._lut_gi_transformer.outPos0  # this will be copied later
                            bins_y = self._lut_gi_transformer.outPos1
                    else:
                        I, bins_x, bins_y, sum, count = \
                            self._lut_gi_transformer.integrate(
                                data, dark=dark, flat=flat,
                                solidAngle=solidangle,
                                dummy=dummy, delta_dummy=delta_dummy,
                                polarization=polarization)

                        # CSR HERE
        csr_transform = """
        if (I  is None) and ("csr" in method):
            logger.debug("in csr")
            mask_crc = None
            with self._lut_sem:
                reset = None
                if self._csr_transformer is None:
                    reset = "init"
                    if mask is None:
                        mask = self.detector.mask
                        mask_crc = self.detector._mask_crc
                    else:
                        mask_crc = crc32(mask)
                if (not reset) and safe:
                    if mask is None:
                        mask = self.detector.mask
                        mask_crc = self.detector._mask_crc
                    else:
                        mask_crc = crc32(mask)
                    if self._csr_gi_transformer.unit != unit:
                        reset = "unit changed"
                    if self._csr_gi_transformer.bins != npt:
                        reset = "number of points changed"
                    if self._csr_gi_transformer.size != data.size:
                        reset = "input image size changed"
                    if (mask is not None) and \
                            (not self._csr_gi_transformer.check_mask):
                        reset = "mask but CSR was without mask"
                    elif (mask is None) and \
                            (self._csr_gi_transformer.check_mask):
                        reset = "no mask but CSR has mask"
                    elif (mask is not None) and \
                            (self._csr_gi_transformer.mask_checksum != \
                            mask_crc):
                        reset = "mask changed"
                    if (x_range is None) and not \
                            self._csr_gi_transformer.default_ipl:
                        reset = "x_range was defined in CSR"
                    elif (x_range is not None) and \
                            (self._csr_gi_transformer.pos0Range != \
                            (min(x_range), max(x_range) * EPS32)):
                        reset = ("x_range is defined"
                                 " but not the same as in CSR")
                    if (y_range is None) and not \
                            self._csr_gi_transformer.default_opl:
                        reset = ("y_range not defined and"
                                 " CSR had y_range defined")
                    elif (y_range is not None) and \
                            (self._csr_gi_transformer.pos1Range != \
                            (min(y_range), max(y_range) * EPS32)):
                        reset = ("y_range requested and"
                                 " CSR's y_range don't match")
                error = False
                if reset:
                    logger.info("gi.transform_image: resetting CSR transformer"
                                " because %s" % reset)
                    if "no" in method:
                        split = "no"
                    elif "full" in method:
                        split = "full"
                    else:
                        split = "bbox"
                    try:
                        sef._csr_gi_transformer = self.gi_setup_CSR(
        """

        if (I is None) and ("splitpix" in method):
            if splitPixel is None:
                logger.warning("splitPixel is not available;"
                               " falling back on default method")
                method = self.DEFAULT_METHOD
            else:
                logger.debug(
                    "pygix.Transform.transform_image uses SplitPixel implementation")
                pos = self.giarray_from_unit(shape, process, "corner", unit)

                I, bins_x, bins_y, sum, count = splitPixel.fullSplit2D(
                    pos=pos,
                    weights=data,
                    bins=npt,
                    pos0Range=x_range,
                    pos1Range=y_range,
                    dummy=dummy,
                    delta_dummy=delta_dummy,
                    mask=mask, dark=dark, flat=flat,
                    solidangle=solidangle,
                    polarization=polarization)

        if (I is None) and ("bbox" in method):
            if splitBBox is None:
                logger.warning(
                    "splitBBox is not available; pygix.Transform.transform_image"
                    " falling back on cython histogram method")
                method = "cython"
            else:
                logger.debug(
                    "pygix.Transform.transform_image uses BBox implementation")
                pos1, pos0 = self.giarray_from_unit(shape, process, "center",
                                                    unit)
                dpos1, dpos0 = self.giarray_from_unit(shape, process, "delta",
                                                      unit)

                I, bins_x, bins_y, sum, count = splitBBox.histoBBox2d(
                    weights=data,
                    pos0=pos0,
                    delta_pos0=dpos0,
                    pos1=pos1,
                    delta_pos1=dpos1,
                    bins=npt,
                    pos0Range=x_range,
                    pos1Range=y_range,
                    dummy=dummy,
                    delta_dummy=delta_dummy,
                    mask=mask, dark=dark, flat=flat,
                    solidangle=solidangle,
                    polarization=polarization,
                    allow_pos0_neg=True)

        if I is None:
            # Common part for np and Cython
            logger.debug(
                "pygix.Transform.transform_image uses cython implementation")
            npt = npt[1], npt[0]
            data = data.astype(np.float32)
            mask = self.make_mask(data, mask, dummy, delta_dummy, mode="np")
            yarray, xarray = self.giarray_from_unit(shape, process, "center",
                                                    unit)

            if x_range is not None:
                mask *= (xarray >= min(x_range)) * (xarray <= max(x_range))
            else:
                x_range = [xarray.min(), xarray.max() * EPS32]

            if y_range is not None:
                mask *= (yarray >= min(y_range)) * (yarray <= max(y_range))
            else:
                y_range = [yarray.min(), yarray.max() * EPS32]

            if variance is not None:
                variance = variance[mask]

            if dark is not None:
                data -= dark

            if flat is not None:
                data /= flat

            if polarization is not None:
                data /= polarization

            if solidangle is not None:
                data /= solidangle

            data = data[mask]
            xarray = xarray[mask]
            yarray = yarray[mask]

            if "cython" in method:
                if histogram is None:
                    logger.warning(
                        "Cython histogram is not available; pygix.Transform.transform_image"
                        " falling back on np histogram")
                    method = "np"
                else:
                    I, bins_y, bins_x, sum, count = histogram.histogram2d(
                        pos0=yarray, pos1=xarray,
                        weights=data, bins=npt,
                        split=False,
                        empty=dummy if dummy is not None else self._empty)

        if I is None:
            method = 'np'
            logger.debug(
                "pygix.Transform.transform_image uses np implementation")
            count, b, c = np.histogram2d(yarray, xarray, npt,
                                         range=[y_range, x_range])
            bins_y = (b[1:] + b[:-1]) / 2.0
            bins_x = (c[1:] + c[:-1]) / 2.0

            count1 = np.maximum(1, count)
            sum, b, c = np.histogram2d(yarray, xarray, npt, weights=data,
                                       range=[y_range, x_range])
            I = sum / count1

        qmask = '''
        if method not in ['np', 'cython']:
            if ('q' in str(unit)) and (self.useqx):
                if self._transformedmask is None:
                    self.make_transformedmask(process, shape, npt, bins_x, bins_y)
                if dummy is None:
                    dummy = 0
            
                I[self._transformedmask == 0] = dummy
                if sigma is not None:
                    sigma[self._transformedmask == 0] = dummy
        #'''
        bins_x *= pos_scale
        if 'polar' in process:
            bins_y = (bins_y - pi) * 180.0 / pi
        else:
            bins_y *= pos_scale

        if normalization_factor:
            I /= normalization_factor

        self._last_method = method

        self.save2D(filename, I, bins_x, bins_y, sigma,
                    dark is not None, flat is not None,
                    polarization_factor, normalization_factor)

        if all:
            res = {"I": I,
                   "in_plane": bins_x,
                   "out_plane": bins_y,
                   "count": count,
                   "sum": sum}
            if sigma is not None:
                res["sigma"] = sigma
        else:
            if sigma is not None:
                res = I, bins_x, bins_y, sigma
            else:
                res = I, bins_x, bins_y
        return res

    def profile_sector(self, data, npt,
                       filename=None, correctSolidAngle=False,
                       variance=None, error_model=None,
                       chi_pos=None, chi_width=None, radial_range=None,
                       mask=None, dummy=None, delta_dummy=None,
                       polarization_factor=None, dark=None, flat=None,
                       method="splitpix", unit=grazing_units.Q,
                       normalization_factor=None):
        """
        Sector integration. The function takes chi_pos (the centre angle of
        the sector and chi_width (the angular opening of the sector). Chi = 0 is
        defined as along the surface normal, e.g., out-of-plane (qz) line
        profile would be at chi = 0 with the angular opening of the sector given
        by chi_width.

        Args:
            data:
            npt:
            filename:
            correctSolidAngle:
            variance:
            error_model:
            chi_pos:
            chi_width:
            radial_range:
            mask:
            dummy:
            delta_dummy:
            polarization_factor:
            dark:
            flat:
            method:
            unit:
            normalization_factor:

        Returns:

        """
        if chi_width is not None:
            azim_min = chi_pos - (chi_width / 2.0)
            azim_max = chi_pos + (chi_width / 2.0)
            azimuth_range = (azim_min, azim_max)
        else:
            azimuth_range = None

        out = self.integrate_1d(
            data, npt, process="sector",
            filename=filename,
            correctSolidAngle=correctSolidAngle,
            variance=variance,
            error_model=error_model,
            p0_range=radial_range,
            p1_range=azimuth_range,
            mask=mask,
            dummy=dummy,
            delta_dummy=delta_dummy,
            polarization_factor=polarization_factor,
            dark=dark, flat=flat,
            method=method,
            unit=unit,
            normalization_factor=normalization_factor
        )
        return out

    def profile_chi(self, data, npt,
                    filename=None, correctSolidAngle=False,
                    variance=None, error_model=None,
                    radial_pos=None, radial_width=None, chi_range=None,
                    mask=None, dummy=None, delta_dummy=None,
                    polarization_factor=None, dark=None, flat=None,
                    method="bbox", unit=grazing_units.Q,
                    normalization_factor=None):
        """
        Chi integration wrapper for integrate_1d.
        """
        rad_min = radial_pos - (radial_width / 2.0)
        rad_max = radial_pos + (radial_width / 2.0)
        radial_range = (rad_min, rad_max)

        out = self.integrate_1d(
            data, npt, process="chi",
            filename=filename,
            correctSolidAngle=correctSolidAngle,
            variance=variance,
            error_model=error_model,
            p0_range=chi_range,
            p1_range=radial_range,
            mask=mask,
            dummy=dummy,
            delta_dummy=delta_dummy,
            polarization_factor=polarization_factor,
            dark=dark,
            flat=flat,
            method=method,
            unit=unit,
            normalization_factor=normalization_factor
        )

        # Inconsitency, for 2D this is in the main function not a wrapper because
        # it handles reciprocal and polar. Maybe it is best to leave this here

        if ("q" in str(unit).lower()) and (self.useqx is True) and (
            method not in ["np", "cython"]):
            I = out[0]
            chi = out[1]

            if self._chimask is None:
                self.make_chimask(rad_max, chi)

            if dummy is None:
                dummy = np.nan

            I[self._chimask == 0] = dummy

            if len(out) == 3:
                sigma = out[2]
                sigma[self._chimask == 0] = dummy
                out = (I, chi, sigma)
            else:
                out = (I, chi)
        return out

    def profile_op_box(self, data, npt,
                       filename=None, correctSolidAngle=False,
                       variance=None, error_model=None,
                       ip_pos=0.0, ip_width=30.0, op_range=None,
                       mask=None, dummy=None, delta_dummy=None,
                       polarization_factor=None, dark=None, flat=None,
                       method="bbox", unit=grazing_units.Q,
                       normalization_factor=None):
        """
        Out-of-plane box integration wrapper for integrate_1d.

        """
        ip_min = ip_pos - (ip_width / 2.0)
        ip_max = ip_pos + (ip_width / 2.0)
        ip_range = (ip_min, ip_max)

        out = self.integrate_1d(
            data, npt,
            process="opbox",
            filename=filename,
            correctSolidAngle=correctSolidAngle,
            variance=variance,
            error_model=error_model,
            p0_range=op_range,
            p1_range=ip_range,
            mask=mask,
            dummy=dummy,
            delta_dummy=delta_dummy,
            polarization_factor=polarization_factor,
            dark=dark,
            flat=flat,
            method=method,
            unit=unit,
            normalization_factor=normalization_factor
        )
        return out

    def profile_ip_box(self, data, npt,
                       filename=None, correctSolidAngle=False,
                       variance=None, error_model=None,
                       op_pos=0.0, op_width=30.0, ip_range=None,
                       mask=None, dummy=None, delta_dummy=None,
                       polarization_factor=None, dark=None, flat=None,
                       method="bbox", unit=grazing_units.Q,
                       normalization_factor=None):
        """
        Out-of-plane box integration wrapper for integrate_1d.

        """
        op_min = op_pos - (op_width / 2.0)
        op_max = op_pos + (op_width / 2.0)
        op_range = (op_min, op_max)

        out = self.integrate_1d(
            data, npt,
            process="ipbox",
            filename=filename,
            correctSolidAngle=correctSolidAngle,
            variance=variance,
            error_model=error_model,
            p0_range=ip_range,
            p1_range=op_range,
            mask=mask,
            dummy=dummy,
            delta_dummy=delta_dummy,
            polarization_factor=polarization_factor,
            dark=dark,
            flat=flat,
            method=method,
            unit=unit,
            normalization_factor=normalization_factor
        )
        return out

    def integrate_1d(self, data, npt, process="sector",
                     filename=None,
                     correctSolidAngle=False,
                     variance=None, error_model=None,
                     p0_range=None, p1_range=None,
                     mask=None, dummy=None, delta_dummy=None,
                     polarization_factor=None, dark=None, flat=None,
                     method="bbox", unit=grazing_units.Q,
                     safe=True, normalization_factor=None):
        """
        Calculate azimuthally integrated 1d curve in q(nm^-1) by 
        efault. Based on pyFAI integrate1d. Multi algorithm 
        implementation (tries to be bullet proof).

        Args:
            data (ndarray): 2D array from detector (raw image).
            npt (int): Number of points in output data.
            filename (str): Output filename in 2/3 column ascii format.
            correctSolidAngle (bool): Correct for solid angle of each pixel.
            variance : ndarray
                Array containing the variance of the data. If not
                available, no error propagation is done.
            error_model : str
                When variance is unknown, an error model can be given:
                "poisson" (variance = I), "azimuthal" (variance =
                (I-<I>)^2).
            p0_range : (float, float), optional
                The lower and upper unit of the radial unit. If not
                provided, range is simply (data.min(), data.max()). Values
                outside the range are ignored.
            p1_range : (float, float), optional
                The lower and upper range of the azimuthal angle in degree.
                If not provided, range is simply (data.min(), data.max()).
                Values outside the range are ignored.
            mask : ndarray
                Masked pixel array (same size as image) with 1 for masked
                pixels and 0 for valid pixels.
            dummy : float
                Value for dead/masked pixels.
            delta_dummy : float
                Precision for dummy value
            polarization_factor : float
                Polarization factor between -1 and +1. 0 for no correction.
            dark : ndarray
                Dark current image.
            flat : ndarray
                Flat field image.
            method : str
                Integration method. Can be "np", "cython", "bbox",
                "splitpix", "lut" or "lut_ocl" (if you want to go on GPU).
            unit : str
                Radial units. Can be "2th_deg", "2th_rad", "q_nm^-1" or
                "q_A^-1" (TTH_DEG, TTH_RAD, Q_NM, Q_A).
            safe : bool
                Do some extra check to ensure LUT is still valid. False is
                faster.
            normalization_factor : float
                Value of a normalization monitor.

        Returns:
            qAxis, I : 2-tuple of ndarrays
                Radial bins and integrated intensity.

            or

            qAxis, I, sigma : 3-tuple of ndarrays
                Radial bins, integrated intensity and error.
        """
        method = method.lower()
        unit = grazing_units.to_unit(unit)
        pos0_scale = 1.0  # nota we need anyway to make a copy !

        if mask is None:
            mask = self.mask

        shape = data.shape
        pos0_scale = unit.scale

        if p0_range:
            if process == "chi":
                p0_range = tuple([np.deg2rad(i) + pi for i in p0_range])
            else:
                p0_range = tuple([i / pos0_scale for i in p0_range])

        if variance is not None:
            assert variance.size == data.size
        elif error_model:
            error_model = error_model.lower()
            if error_model == "poisson":
                variance = np.ascontiguousarray(data, np.float32)

        if p1_range is not None:
            if process == "sector":
                p1_range = tuple([np.deg2rad(i) + pi for i in p1_range])
            else:
                p1_range = tuple([i / pos0_scale for i in p1_range])

        if correctSolidAngle:
            solidangle = self.solidAngleArray(shape, correctSolidAngle)
        else:
            solidangle = None

        if polarization_factor is None:
            polarization = None
        else:
            polarization = self.polarization(shape, float(polarization_factor))

        if dark is None:
            dark = self.darkcurrent

        if flat is None:
            flat = self.flatfield

        I = None
        sigma = None

        if (I is None) and ("lut" in method):
            mask_crc = None
            with self._lut_sem:
                reset = None
                if self._lut_gi_integrator is None:
                    reset = "init"
                    if mask is None:
                        mask = self.detector.mask
                        mask_crc = self.detector._mask_crc
                    else:
                        mask_crc = crc32(mask)
                if (not reset) and safe:
                    if mask is None:
                        mask = self.detector.mask
                        mask_crc = self.detector._mask_crc
                    else:
                        mask_crc = crc32(mask)
                    if self._lut_gi_integrator.unit != unit:
                        reset = "unit changed"
                    if self._lut_gi_integrator.bins != npt:
                        reset = "number of points changed"
                    if self._lut_gi_integrator.size != data.size:
                        reset = "input image size changed"
                    if (mask is not None) and (
                    not self._lut_gi_integrator.check_mask):
                        reset = "mask but LUT was without mask"
                    elif (mask is None) and self._lut_gi_integrator.check_mask:
                        reset = "no mask but LUT has mask"
                    elif (mask is not None) and (
                        self._lut_gi_integrator.mask_checksum != mask_crc):
                        reset = "mask changed"
                    if (
                        p0_range is None) and not self._lut_gi_integrator.default_p0:
                        reset = "p0_range was defined in LUT"
                    elif (p0_range is not None) and \
                            (self._lut_gi_integrator.pos0Range != (
                            min(p0_range), max(p0_range) * EPS32)):
                        reset = "p0_range is defined but not the same as in LUT"
                    if (
                        p1_range is None) and not self._lut_gi_integrator.default_p1:
                        reset = "p1_range not defined and LUT had p1_range defined"
                    elif (p1_range is not None) and \
                            (self._lut_gi_integrator.pos1Range != (
                            min(p1_range), max(p1_range) * EPS32)):
                        reset = "p1_range requested and LUT's p1_range don't match"
                if reset:
                    logger.info(
                        "pygix.Transform.integrate_1d: Resetting integrator because %s" % reset)
                    try:
                        lut_giIntegrator, def_p0, def_p1 = \
                            self.gi_setup_lut(shape, npt,
                                              process, mask,
                                              p0_range,
                                              p1_range,
                                              mask_checksum=mask_crc,
                                              unit=unit)

                        self._lut_gi_integrator = lut_giIntegrator
                        self._lut_gi_integrator.default_p0 = def_p0
                        self._lut_gi_integrator.default_p1 = def_p1
                    except MemoryError:  # LUT method is hungry...
                        logger.warning(
                            "MemoryError: pygix.Transform.integrate_1d falling back on"
                            " forward implementation")
                        self._ocl_lut_gi_integrator = None
                        self._lut_gi_integrator = None
                        gc.collect()
                        method = "splitbbox"
                if self._lut_gi_integrator:
                    if ("ocl" in method) and ocl_azim_lut:
                        with self._ocl_lut_sem:
                            if "," in method:
                                c = method.index(",")
                                platformid = int(method[c - 1])
                                deviceid = int(method[c + 1])
                                devicetype = "all"
                            elif "gpu" in method:
                                platformid = None
                                deviceid = None
                                devicetype = "gpu"
                            elif "cpu" in method:
                                platformid = None
                                deviceid = None
                                devicetype = "cpu"
                            else:
                                platformid = None
                                deviceid = None
                                devicetype = "all"
                            if (self._ocl_lut_gi_integrator is None) or \
                                    (self._ocl_lut_gi_integrator.on_device[
                                         "lut"] != \
                                             self._lut_gi_integrator.lut_checksum):
                                self._ocl_lut_gi_integrator = \
                                    ocl_azim_lut.OCL_LUT_Integrator(
                                        self._lut_gi_integrator.lut,
                                        self._lut_gi_integrator.size,
                                        devicetype=devicetype,
                                        platformid=platformid,
                                        deviceid=deviceid,
                                        checksum= \
                                            self._lut_gi_integrator.lut_checksum)

                            I, _, _ = self._ocl_lut_gi_integrator.integrate(
                                data, dark=dark, flat=flat,
                                solidAngle=solidangle,
                                solidAngle_checksum=self._dssa_crc,
                                dummy=dummy,
                                delta_dummy=delta_dummy,
                                polarization=polarization,
                                polarization_checksum= \
                                    self._polarization_crc)

                            qAxis = self._lut_gi_integrator.outPos  # this will be copied later
                            if error_model == "azimuthal":
                                variance = (data - self.calcfrom1d(qAxis * \
                                                                   pos0_scale,
                                                                   I,
                                                                   dim1_unit=unit)) ** 2
                            if variance is not None:
                                var1d, a, b = self._ocl_lut_gi_integrator.integrate(
                                    variance,
                                    solidAngle=None,
                                    dummy=dummy,
                                    delta_dummy=delta_dummy)
                                sigma = np.sqrt(a) / np.maximum(b, 1)
                    else:
                        qAxis, I, a, b = self._lut_gi_integrator.integrate(
                            data, dark=dark, flat=flat,
                            solidAngle=solidangle,
                            dummy=dummy,
                            delta_dummy=delta_dummy,
                            polarization=polarization)

                        if error_model == "azimuthal":
                            variance = (data - self.calcfrom1d(qAxis * \
                                                               pos0_scale, I,
                                                               dim1_unit=unit)) ** 2
                        if variance is not None:
                            _, var1d, a, b = self._lut_gi_integrator.integrate(
                                variance,
                                solidAngle=None,
                                dummy=dummy,
                                delta_dummy=delta_dummy)
                            sigma = np.sqrt(a) / np.maximum(b, 1)

        if (I is None) and ("splitpix" in method):
            if splitPixel is None:
                logger.warning(
                    "SplitPixel is not available, pygix.Transform.integrate_1d"
                    " falling back on splitbbox histogram !")
                method = "splitbbox"
            else:
                logger.debug(
                    "pygix.Transform.integrate_1d uses SplitPixel implementation")
                pos = self.giarray_from_unit(shape, process, "corner", unit)

                qAxis, I, a, b = splitPixel.fullSplit1D(
                    pos=pos,
                    weights=data,
                    bins=npt,
                    pos0Range=p0_range,
                    pos1Range=p1_range,
                    dummy=dummy,
                    delta_dummy=delta_dummy,
                    mask=mask,
                    dark=dark,
                    flat=flat,
                    solidangle=solidangle,
                    polarization=polarization)
                if error_model == "azimuthal":
                    variance = (data - self.calcfrom1d(qAxis * \
                                                       pos0_scale, I,
                                                       dim1_unit=unit)) ** 2
                if variance is not None:
                    _, var1d, a, b = splitPixel.fullSplit1D(
                        pos=pos,
                        weights=variance,
                        bins=npt,
                        pos0Range=p0_range,
                        pos1Range=p1_range,
                        dummy=dummy,
                        delta_dummy=delta_dummy,
                        mask=mask)
                    sigma = np.sqrt(a) / np.maximum(b, 1)

        if (I is None) and ("bbox" in method):
            if splitBBox is None:
                logger.warning(
                    "pyFAI.splitBBox is not available, pygix.Transform.integrate_1d"
                    " falling back on cython histograms")
                method = "cython"
            else:
                logger.debug(
                    "pygix.Transform.integrate_1d uses BBox implementation")
                chi, pos0 = self.giarray_from_unit(shape, process, "center",
                                                   unit)
                dchi, dpos0 = self.giarray_from_unit(shape, process, "delta",
                                                     unit)

                qAxis, I, a, b = splitBBox.histoBBox1d(
                    weights=data,
                    pos0=pos0,
                    delta_pos0=dpos0,
                    pos1=chi,
                    delta_pos1=dchi,
                    bins=npt,
                    pos0Range=p0_range,
                    pos1Range=p1_range,
                    dummy=dummy,
                    delta_dummy=delta_dummy,
                    mask=mask,
                    dark=dark,
                    flat=flat,
                    solidangle=solidangle,
                    polarization=polarization)
                if error_model == "azimuthal":
                    variance = (data - self.calcfrom1d(qAxis * pos0_scale, I,
                                                       dim1_unit=unit)) ** 2
                if variance is not None:
                    _, var1d, a, b = splitBBox.histoBBox1d(
                        weights=variance,
                        pos0=pos0,
                        delta_pos0=dpos0,
                        pos1=chi,
                        delta_pos1=dchi,
                        bins=npt,
                        pos0Range=p0_range,
                        pos1Range=p1_range,
                        dummy=dummy,
                        delta_dummy=delta_dummy,
                        mask=mask)
                    b[b == 0] = 1
                    sigma = np.sqrt(a) / b

        if I is None:
            # Common part for  np and Cython
            data = data.astype(np.float32)
            mask = self.make_mask(data, mask, dummy, delta_dummy, mode="np")
            pos1, pos0 = self.giarray_from_unit(shape, process, "center", unit)

            if p0_range is not None:
                mask *= (pos0 >= min(p0_range)) * \
                        (pos0 <= max(p0_range))
            else:
                p0_range = [pos0.min(), pos0.max() * EPS32]
            if p1_range is not None:
                mask *= (pos1 >= min(p1_range)) * \
                        (pos1 <= max(p1_range))
            else:
                p1_range = [pos1.min(), pos1.max() * EPS32]

            mask = np.where(mask)

            if dark is not None:
                data -= dark

            if flat is not None:
                data /= flat

            if polarization is not None:
                data /= polarization

            if solidangle is not None:
                data /= solidangle

            if variance is not None:
                variance = variance[mask]

            data = data[mask]
            pos0 = pos0[mask]
            pos1 = pos1[mask]

            if "cython" in method:
                if histogram is not None:
                    logger.debug("giIntegrate1d uses cython implementation")
                    if dummy is None:
                        dummy = 0

                    qAxis, I, a, b = histogram.histogram(pos=pos0,
                                                         weights=data,
                                                         bins=npt,
                                                         pixelSize_in_Pos=0,
                                                         empty=dummy)
                    if error_model == "azimuthal":
                        variance = (data - self.calcfrom1d(
                            qAxis * pos0_scale,
                            I, dim1_unit=unit,
                            correctSolidAngle=False)[mask]) ** 2
                    if variance is not None:
                        _, var1d, a, b = histogram.histogram(pos=pos0,
                                                             weights=variance,
                                                             bins=npt,
                                                             pixelSize_in_Pos=1,
                                                             empty=dummy)
                        sigma = np.sqrt(a) / np.maximum(b, 1)
                else:
                    logger.warning("pyFAI.histogram is not available,"
                                   " falling back on np")
                    method = "np"

        if I is None:
            logger.debug("pygix.Transform.integrate_1d uses np implementation")
            method = "np"
            ref, b = np.histogram(pos0, npt, range=p0_range)
            qAxis = (b[1:] + b[:-1]) / 2.0
            count = np.maximum(1, ref)
            val, b = np.histogram(pos0, npt,
                                  weights=data, range=p0_range)
            if error_model == "azimuthal":
                variance = (data - self.calcfrom1d(
                    qAxis * pos0_scale,
                    I, dim1_unit=unit,
                    correctSolidAngle=False)[mask]) ** 2
            if variance is not None:
                var1d, b = np.histogram(pos0, npt,
                                        weights=variance,
                                        range=p0_range)
                sigma = np.sqrt(var1d) / count
            I = val / count

        if process == "chi":
            qAxis = np.rad2deg(qAxis - pi)
        elif pos0_scale:
            qAxis = qAxis * pos0_scale

        if normalization_factor:
            I /= normalization_factor
            if sigma is not None:
                sigma /= normalization_factor

        self.save1D(filename, qAxis, I, sigma, unit,
                    dark is not None, flat is not None,
                    polarization_factor=polarization_factor,
                    normalization_factor=normalization_factor)

        if sigma is not None:
            return I, qAxis, sigma
        else:
            return I, qAxis

    def save2D(self, filename, intensity, x_scale, y_scale, error=None,
               has_dark=False, has_flat=False,
               polarization_factor=None, normalization_factor=None):
        """

        Args:
            filename:
            intensity:
            x_scale:
            y_scale:
            error:
            has_dark:
            has_flat:
            polarization_factor:
            normalization_factor:

        Returns:

        """
        if not filename:
            return
        writer = io.Writer(self, None)
        writer.save2D(filename, intensity, x_scale, y_scale, error,
                      has_dark, has_flat, polarization_factor,
                      normalization_factor)

    def save1D(self, filename, x_scale, intensity, error=None,
               x_unit=grazing_units.Q,
               has_dark=False, has_flat=False, polarization_factor=None,
               normalization_factor=None):
        """
        @param filename: the filename used to save the 1D integration
        @type filename: str
        @param x_scale: the x coordinates of the integrated curve
        @type x_scale: numpy.ndarray
        @param intensity: The integrated intensity
        @type intensity: numpy.mdarray
        @param error: the error bar for each intensity
        @type error: numpy.ndarray or None
        @param x_unit: the unit of the dim1 array
        @type x_unit: pyFAI.units.Unit
        @param has_dark: save the darks filenames (default: no)
        @type has_dark: bool
        @param has_flat: save the flat filenames (default: no)
        @type has_flat: bool
        @param polarization_factor: the polarization factor
        @type polarization_factor: float
        @param normalization_factor: the monitor value
        @type normalization_factor: float)

        This method save the result of a 1D integration.
        """
        if not filename:
            return
        writer = io.Writer(self, None)
        writer.save1D(filename, x_scale, intensity, error, x_unit,
                      has_dark, has_flat, polarization_factor,
                      normalization_factor)

    # --------------------------------------------------------------------------
    #   Some properties
    # --------------------------------------------------------------------------
    #       All taken from pyFAI.azimuthalIntegrator.AzimuthalIntegrator
    # TODO !!! Get rid of all of these. Surely I don't need to have them here.

    def set_maskfile(self, maskfile):
        self.detector.set_maskfile(maskfile)

    def get_maskfile(self):
        return self.detector.get_maskfile()

    maskfile = property(get_maskfile, set_maskfile)

    def set_mask(self, mask):
        self.detector.set_mask(mask)

    def get_mask(self):
        return self.detector.get_mask()

    mask = property(get_mask, set_mask)

    def set_darkcurrent(self, dark):
        self._darkcurrent = dark
        if dark is not None:
            self._darkcurrent_crc = crc32(dark)
        else:
            self._darkcurrent_crc = None

    def get_darkcurrent(self):
        return self._darkcurrent

    darkcurrent = property(get_darkcurrent, set_darkcurrent)

    def set_flatfield(self, flat):
        self._flatfield = flat
        if flat is not None:
            self._flatfield_crc = crc32(flat)
        else:
            self._flatfield_crc = None

    def get_flatfield(self):
        return self._flatfield

    flatfield = property(get_flatfield, set_flatfield)

    def set_darkfiles(self, files=None, method="mean"):
        """
        Set the dark current from one or mutliple files, avaraged
        according to the method provided.

        Parameters
        ----------
        files : str or list(str) or None
            File(s) used to compute the dark current image.
        method : str 
            Method used to compute the dark. Can be "mean" or "median".
        """
        if type(files) in types.StringTypes:
            files = [i.strip() for i in files.split(",")]
        elif not files:
            files = []
        if len(files) == 0:
            self.set_darkcurrent(None)
        elif len(files) == 1:
            self.set_darkcurrent(
                fabio.open(files[0]).data.astype(np.float32))
            self.darkfiles = files[0]
        else:
            self.set_darkcurrent(pyFAI.utils.averageImages(
                files, filter_=method,
                format=None, threshold=0))
            self.darkfiles = "%s(%s)" % (method, ",".join(files))

    def set_flatfiles(self, files, method="mean"):
        """
        Set the flat field from one or mutliple files, averaged
        according to the method provided.

        Parameters
        ----------
        files : str or list(str) or None
            File(s) used to compute the flat field image.
        method : str 
            Method used to compute the flat. Can be "mean" or "median".
        """
        if type(files) in types.StringTypes:
            files = [i.strip() for i in files.split(",")]
        elif not files:
            files = []
        if len(files) == 0:
            self.set_flatfield(None)
        elif len(files) == 1:
            self.set_flatfield(fabio.open(files[0]).data.astype(np.float32))
            self.flatfiles = files[0]
        else:
            self.set_flatfield(pyFAI.utils.averageImages(
                files, filter_=method,
                format=None, threshold=0))
            self.flatfiles = "%s(%s)" % (method, ",".join(files))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

"""Unique place where the version number is defined."""

__authors__ = ["Jérôme Kieffer"]
__license__ = "MIT"
__date__ = "17/03/2016"


# Do not copy into the source folder !

from collections import namedtuple
_version_info = namedtuple(
    "version_info", ["major", "minor", "micro", "releaselevel", "serial"])

MAJOR = 0
MINOR = 1
MICRO = 4
RELEV = "dev"  # <16
SERIAL = 0  # <16

version_info = _version_info(MAJOR, MINOR, MICRO, RELEV, SERIAL)

strictversion = version = "%d.%d.%d" % version_info[:3]

RELEASE_LEVEL_VALUE = {"dev": 0,
                       "alpha": 10,
                       "beta": 11,
                       "rc": 12,
                       "final": 15}

if version_info.releaselevel != "final":
    version += "-%s%s" % version_info[-2:]
    prerel = "a" if RELEASE_LEVEL_VALUE.get(version_info[3], 0) < 10 else "b"
    if prerel not in "ab":
        prerel = "a"
    strictversion += prerel + str(version_info[-1])

hexversion = version_info[4]
hexversion |= RELEASE_LEVEL_VALUE.get(version_info[3], 0) * 1 << 4
hexversion |= version_info[2] * 1 << 8
hexversion |= version_info[1] * 1 << 16
hexversion |= version_info[0] * 1 << 24

if __name__ == '__main__':
    print(version)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

__author__ = "Thomas Dane, Jerome Kieffer"
__contact__ = "thomasgdane@gmail.com"
__license__ = "GPLv3+"
__copyright__ = "ESRF - The European Synchrotron, Grenoble, France"
__date__ = "18/11/2014"
__status__ = "Development"
__docformat__ = "restructuredtext"

import logging
logger = logging.getLogger("pygix.grazing_units")
from numpy import pi
import types
hc = 12.398419292004204


class Enum(dict):
    """
    Simple class half way between a dict and a class, behaving as an enum
    """
    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError

    def __repr__(self, *args, **kwargs):
        if "REPR" in self:
            return self["REPR"]
        else:
            return dict.__repr__(self, *args, **kwargs)

UNDEFINED = Enum(REPR='?')

somethoughts="""
ANGULAR = Enum(REPR="angular",
               x_unit="2theta_f",
               y_unit="alpha_f",
               center="gia_center_array",
               corner="gia_corner_array",
               delta="gia_delta_array",
               scale={'deg' : (180.0/pi),
                      'rad' : 1.0}

RECIPROCAL = Enum(REPR="reciprocal",
                  x_unit="qxy",
                  y_unit="qz",
                  center="giq_center_array",
                  corner="giq_corner_array",
                  delta="giq_delta_array",
                  scale={'nm' : 100.0),
                         'A' : 10.0}

POLAR = Enum(REPR="angular",
             x_unit="q",
             y_unit="chi",
             center="absq_center_array",
             corner="absq_corner_array",
             delta="absq_delta_array",
             scale={'deg' : (180.0/pi),
                    'rad' : 1.0}

SECTOR

CHI

IPBOX

OPBOX


TTH = Enum(REPR="angular",
           x_unit="2theta_f",
           y_unit="alpha_f",
           center="gia_center_array",
           corner="gia_corner_array",
           delta="gia_delta_array",
           abs_center="absa_center_array",
           abs_corner="absa_corner_array",
           abs_delta="absa_delta_array",
           scale={'deg' : (180.0/pi),
                  'rad' : 1.0}
Q = Enum(REPR='reciprocal',
         x_unit='qxy

"""

TTH_DEG = TTH = Enum(REPR="2theta_deg",
                     IPL_UNIT="2theta_f_deg",
                     OPL_UNIT="alpha_f_deg",
                     center="gia_center_array",
                     corner="gia_corner_array",
                     delta="gia_delta_array",
                     abs_center="absa_center_array",
                     abs_corner="absa_corner_array",
                     abs_delta="absa_delta_array",
                     scale=180.0/pi)

TTH_RAD = Enum(REPR="2theta_rad",
               IPL_UNIT="2theta_f_rad",
               OPL_UNIT="alpha_f_rad",  
               center="gia_center_array",
               corner="gia_corner_array",
               delta="gia_delta_array",
               abs_center="absa_center_array",
               abs_corner="absa_corner_array",
               abs_delta="absa_delta_array",     
               scale=1.0)

Q_NM = Q = Enum(REPR="q_nm^-1",
                IPL_UNIT="qxy_nm^-1",
                OPL_UNIT="qz_nm^-1",
                center="giq_center_array",
                corner="giq_corner_array",
                delta="giq_delta_array",
                abs_center="absq_center_array",
                abs_corner="absq_corner_array",
                abs_delta="absq_delta_array",
                scale=100.0)

Q_A = Enum(REPR="q_A^-1",
           IPL_UNIT="qxy_A^-1",
           OPL_UNIT="qz_A^-1",
           center="giq_center_array",
           corner="giq_corner_array",
           delta="giq_delta_array",
           abs_center="absq_center_array",
           abs_corner="absq_corner_array",
           abs_delta="absq_delta_array",
           scale=10)

GI_UNITS = (TTH_DEG, TTH_RAD, Q_NM, Q_A)

def to_unit(obj):
    giUnit = None
    if isinstance(obj, str):
        for one_unit in GI_UNITS:
            if one_unit.REPR == obj:
                giUnit = one_unit
                break
    elif obj.__class__.__name__.split(".")[-1] == "Enum":
        giUnit = obj
    if giUnit is None:
        logger.error("Unable to recognize this type unit '%s' of type %s."
                     " Valid units are 2th_deg, 2th_rad, q_nm^-1 and q_A^-1" % \
                      (obj, type(obj)))
    return giUnit

def ip_unit(obj):
    ipl_giUnit = None
    if type(obj) in types.StringTypes:
        for one_unit in GI_UNITS:
            if one_unit.IPL_UNIT == obj:
                ipl_giUnit = one_unit
                break
    elif obj.__class__.__name__.split(".")[-1] == "Enum":
        ipl_giUnit = obj
    if ipl_giUnit is None:
        logger.error("Unable to recognize this type unit '%s' of type %s."
                     " Valid units are 2th_deg, 2th_rad, q_nm^-1 and q_A^-1" % \
                     (obj, type(obj)))
    return ipl_giUnit  

def op_unit(obj):
    opl_giUnit = None
    if type(obj) in types.StringTypes:
        for one_unit in GI_UNITS:
            if one_unit.OPL_UNIT == obj:
                opl_giUnit = one_unit
                break
    elif obj.__class__.__name__.split(".")[-1] == "Enum":
        opl_giUnit = obj
    if opl_giUnit is None:
        logger.error("Unable to recognize this type unit '%s' of type %s."
                     " Valid units are 2th_deg, 2th_rad, q_nm^-1 and q_A^-1" % \
                     (obj, type(obj)))
    return opl_giUnit  


# ABS_TTH_DEG = ABS_TTH = Enum(REPR="2th_deg",
#                      center="chi_tth_center_array",
#                      corner="chi_tth_corner_array",
#                      delta="chi_tth_delta_array",
#                      scale=180.0 / pi)

# ABS_TTH_RAD = Enum(REPR="2th_rad",
#                center="chi_tth_center_array",
#                corner="chi_tth_corner_array",
#                delta="chi_tth_delta_array",
#                scale=1.0)

# ABS_Q_NM = ABS_Q = Enum(REPR="q_nm^-1",
#                 center="chi_q_center_array",
#                 corner="chi_q_corner_array",
#                 delta="chi_q_delta_array",
#                 scale=100.0)

# ABS_Q_A = Enum(REPR="q_A^-1",
#            center="chi_q_center_array",
#            corner="chi_q_corner_array",
#            delta="chi_q_delta_array",
#            scale=10.0)

# ABS_UNITS = (ABS_TTH_DEG, ABS_TTH_RAD, ABS_Q_NM, ABS_Q_A)

def absolute_unit(obj):
    abs_giUnit = None
    print(type(obj))
    if type(obj) in types.StringTypes:
        for one_unit in ABS_UNITS:
            if one_unit.REPR == obj:
                abs_giUnit = one_unit
                break
    elif obj.__class__.__name__.split(".")[-1] == "Enum":
        abs_giUnit = obj
    if abs_giUnit is None:
        logger.error("Unable to recognize this type unit '%s' of type %s. "
                     "Valid units are 2th_deg, 2th_rad, q_nm^-1, q_A^-1 and r_mm"\
                     % (obj, type(obj)))
    print(abs_giUnit)
    return abs_giUnit        




#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Processor module for batch processing data reduction.

The user creates a yaml file[1] containing all of the detector and sample
geometry parameters, the input files, output files and the data reduction
parameters. Passing this file to the processor, all input files will be
processed and output data saved if requested. If batch 1D is used, an array
containing the stacked 1D profiles will be returned from the processor.

Usage:

import pygix.process as ppr

processor = ppr.Process('process_recipe.yaml')
out = processor.process()

"""

import sys
import os
from glob import glob
import yaml
import fabio
import pygix
import numpy as np
import time


class Processor(object):
    """
    Class for batch processing data with pygix module. Takes a
    yaml file, which lists geometry, correction files, input data,
    output file names and integration parameters and performs the
    data reduction.
    """

    def __init__(self, yaml_file):
        """
        Initialization. Takes the yaml_file and reads in all parameters, which
        are stored as class attributes. Instatiates pygix.transform.Transform.

        Args:
            yaml_file (string): path to recipe yaml file.
        """
        self.recipe = yaml_file
        with open(yaml_file, 'r') as f:
            self.pars = yaml.load(f)

        try:
            calibration = self.pars['calibration']
        except KeyError:
            raise RuntimeError('calibration data not present in yaml file')
        self._pg = init_pygix(calibration)

        try:
            data = self.pars['data']
        except KeyError:
            raise RuntimeError('data information not present in yaml file')
        self.file_list = get_file_list(data['infiles']['dname'],
                                            data['infiles']['fname'],
                                            data['infiles']['numbers'])

        if 'outfiles' in data.keys():
            raise NotImplementedError('Saving data not yet implemented')

        # self.bkg_dname = data['backfiles']['dname']
        # self.bkg_fname = data['backfiles']['fname']
        # self.bkg_numbers = data['backfiles']['numbers']
        # self.out_dname = data['outfiles']['dname']
        # self.out_fname = data['outfiles']['fname']

        # set basename and extn checks for compressed formats like file.edf.bz2
        # basename, extn = os.path.splitext(self.in_fname)
        # compressed_formats = ['bz2', 'gz']
        # if extn in compressed_formats:
        #     self.basename = os.path.splitext(basename)[0]
        #     self.extension = os.path.splitext(basename)[1] + '.' + extn
        # else:
        #     self.basename = basename
        #     self.extension = extn

        red_methods = ['transform_reciprocal',
                       'transform_polar',
                       'transform_angular',
                       'profile_sector',
                       'profile_chi',
                       'profile_op_box',
                       'profile_ip_box']

        try:
            reduction = self.pars['data_reduction']
        except KeyError:
            raise RuntimeError('reduction parameters not in yaml file')

        self.reduction = reduction.keys()[0]
        if self.reduction not in red_methods:
            raise RuntimeError(('Invalid reduction method: %s \nValid reduction'
                                ' methods: %s') % (self.reduction, red_methods))

        red_kwargs = reduction[self.reduction]

        # set as class attribute and remove if value is None otherwise
        # can interfere with integration if kwarg = None
        self.red_kwargs = {}
        if reduction[self.reduction].__class__ is dict:
            for key in red_kwargs:
                if red_kwargs[key] is not None:
                    self.red_kwargs.update({key: red_kwargs[key]})

            for key in self.red_kwargs:
                if 'range' in key:
                    rng = ''.join(
                        c for c in self.red_kwargs[key] if c not in '()')
                    rng = rng.strip().split(',')
                    self.red_kwargs[key] = tuple(float(x) for x in rng)

        self._reducer = self.set_reducer()

    def set_reducer(self):
        """
        Returns the reduction function (called in __init__ and set as
        class attribute.
        """
        return getattr(self._pg, self.reduction)

    def do_reduction(self, filename):
        """
        Loads the data from filename and returns the reduced data.

        Args:
            filename (string): path to data.

        Returns:
            data (ndarray): reduced data. This can be 1D or 2D depending on
                what was specified in the yaml file.
        """
        data = fabio.open(filename).data
        if os.path.splitext(filename)[1] == '.tif':
            data = np.flipud(data)
        return self._reducer(data, **self.red_kwargs)

    def _process1d(self):
        """ Core function for batch processing data (1D reduction).

        All information is stored as class attributes
        """
        file_list = self.file_list
        n_files = len(file_list)
        out_array = np.zeros((n_files, self.red_kwargs['npt']))

        t_start = time.time()

        for i, f in enumerate(file_list):
            print('processing file: {}/{}\n\t{}'.format(i + 1, n_files, f))

            # THIS WORKS BUT NEEDS TESTING
            # if self.out_fname is not None:
            #    out_fname = rootname+str(i).zfill(nz)+'_p'+self.extension
            #    self.red_kwargs['filename']  = os.path.join(self.out_dname,
            #       out_fname)

            out_array[i, ], x_scale = self.do_reduction(f)

        t_end = time.time() - t_start
        msg = '{} files processed in {} seconds\n'
        msg += '{} seconds per file'
        print(msg.format(n_files, t_end, t_end/n_files))

        y_scale = np.arange(0, n_files)
        out = (out_array, x_scale, y_scale)

        # sv_array = fabio.edfimage.edfimage(out_array)
        # sv_array.write('test.edf')
        return out

    def _process2d(self):
        """
        :return:
        """
        file_list = self.file_list
        n_files = len(file_list)

        t_start = time.time()

        for i, f in enumerate(file_list):
            print('processing file: {}/{}\n\t{}'.format(i + 1, n_files, f))

            # THIS WORKS BUT NEEDS TESTING
            # if self.out_fname is not None:
            #    out_fname = rootname+str(i).zfill(nz)+'_p'+self.extension
            #    self.red_kwargs['filename']  = os.path.join(self.out_dname,
            #       out_fname)
            self.do_reduction(f)

        t_end = time.time() - t_start
        msg = '{} files processed in {} seconds\n'
        msg += '{} seconds per file'
        print(msg.format(n_files, t_end, t_end/n_files))

        return True

    def process(self):
        """
        Main batch processor. Will get file list the iterate over 
        each performing reduction. Will save the data if outfiles 
        have been specified. Will plot all if live_view is True,
        requests user input if interactive is True.
        """
        # CHECKS FOR WRITING DATA
        # if (self.out_fname is not None) and (self.out_dname is None):
        #    raise RuntimeError(('Cannot write reduced data as no reduced '
        #        'directory specified.'))
        # if (self.out_dname is not None) and (self.out_fname is None):
        #    self.out_fname = self.basename.rstrip('0')
        # elif (self.out_fname is not None) and (self.out_dname is not None):
        #    rootname = self.basename.rstrip('0')
        #    nz = len(self.basename)-len(rootname)

        if 'profile' in self.reduction:
            out = self._process1d()
        elif 'transform' in self.reduction:
            out = self._process2d()
        return out


def list_to_indices(index_string):
    """
    Return an integer list from a string representing indices.
    e.g. index_string = '1-3, 5-6, 8-13, 15, 20'
         indices = [1, 2, 3, 5, 6, 8, 9, 10, 11, 12, 13, 15, 20]

    Args:
        index_string (string): condensed string representation of integer list.

    Returns:
        indices (list):
    """
    indices = []
    for s in index_string.split(','):
        if '-' in s:
            first, last = s.split('-')
            for i in range(int(first), int(last) + 1):
                indices.append(i)
        else:
            indices.append(int(s))
    return indices


def indices_to_list(indices):
    """
    Return an abbreviated string representing indices.
    e.g. indices = [1, 2, 3, 5, 6, 8, 9, 10, 11, 12, 13, 15, 20]
         index_string = '1-3, 5-6, 8-13, 15, 20'

    Args:
        indices (list):

    Returns:
        index_string (string): condensed string representation of integer list.
    """
    index_string = ""
    end = start = indices[0]
    for i in range(1, len(indices)):
        if indices[i] == (indices[i - 1] + 1):
            end = indices[i]
        else:
            if start == end:
                index_string += str(start) + ","
            else:
                index_string += str(start) + "-" + str(end) + ","
            start = end = indices[i]
    if start == end:
        index_string += str(start)
    else:
        index_string += str(start) + "-" + str(end)
    return index_string


def get_file_list(dname, fname, numbers=None):
    """
    Takes a directory path, filename format and (optionally) frame numbers and
    returns a list of full file paths.

    Args:
        dname (string): directory path.
        fname (string): basename for images. Can be full filename or can contain
            wildcard ('*', or '0000').
        numbers: (string, list or None): contains information on images with
            fname to be used. If None ('*' or '00...' must be in fname), will
            take all images. Can be an integer list of file numbers or can be a
            string with hyphens representing ranges, e.g., '1, 3, 5, 9-15'.

    Returns:
        file_list (list): list of full paths to data.
    """
    if not os.path.isdir(dname):
        raise IOError('Directory does not exist!\n{}'.format(dname))
    elif (numbers is None) and ('*' not in fname):
        raise IOError(
            'No file numbers provided and no wildcard (*) in filename')

    compressed_formats = ['bz2', 'gz']

    if (numbers is None) and ('*' in fname):
        file_list = sorted(glob(os.path.join(dname, fname)))
    else:
        if '*' in fname:
            fname = fname.replace('*', '{:04d}')
        else:
            basename, extn = os.path.splitext(fname)
            if extn in compressed_formats:
                basename, tmp_extn = os.path.splitext(basename)
                extn = '{}{}'.format(tmp_extn, extn)

            zero_stripped = basename.rstrip('0')
            n_zeros = len(basename) - len(zero_stripped)
            if n_zeros > 0:
                fname = '{}{{:0{}d}}{}'.format(zero_stripped, n_zeros, extn)
            elif '{:0' not in fname:
                raise IOError('bad filename specifier')

        if numbers.__class__ == str:
            numbers = list_to_indices(numbers)

        file_list = []
        for i in numbers:
            in_file = fname.format(i)
            file_list.append(os.path.join(dname, in_file))
    return file_list


def init_pygix(calibration_dict):
    """
    Instantiate a pygix.Transform() class instance.
    All parameters are set from the yaml file.

    Args:
        calibration_dict (dict): dictionary containing all parameters
            for instantiating pygix.Transform().

    Returns:
        pg (object): class instance of pygix.Transform().
    """
    pg = pygix.transform.Transform()
    pg.load(calibration_dict['ponifile'])

    if 'splinefile' in calibration_dict.keys():
        pg.splinefile = calibration_dict['splinefile']
    if 'flatfile' in calibration_dict.keys():
        pg.flatfiles = calibration_dict['flatfile']
    if 'darkfile' in calibration_dict.keys():
        pg.darkfiles = calibration_dict['darkfile']
    if 'maskfile' in calibration_dict.keys():
        pg.maskfile = calibration_dict['maskfile']

    grazing = calibration_dict['grazing_parameters']
    if 'sample_orientation' in grazing.keys():
        pg.sample_orientation = grazing['sample_orientation']
    else:
        pg.sample_orientation = 1

    pg.incident_angle = grazing['incident_angle']
    if 'tilt_angle' in grazing.keys():
        pg.tilt_angle = grazing['tilt_angle']
    return pg


if __name__ == '__main__':
    if len(sys.argv) is not 2:
        print('usage: process.py recipe.yaml')
        sys.exit(0)
    else:
        rp = Processor(sys.argv[1])
        rp.process()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" input/output functions.
"""

from collections import OrderedDict
import fabio
from . import grazing_units


class Writer(object):
    """

    """

    def __init__(self, pg, filename):
        """

        Returns:

        """
        self._pg = pg
        self._filename = filename
        self._hdr_dict = None
        self._header = None
        self._already_written = False

    def make_header_dict(self, has_dark=False, has_flat=False,
                         polarization_factor=None, normalization_factor=None):
        """

        Args:
            hdr (string): string used as comment in the header.
            has_dark (bool): save the file names of dark images.
            has_flat (bool): save the file names of flat images.
            polarization_factor (float):
            normalization_factor (float):

        Returns:
            self._hdr_dict (OrderedDict):
        """
        if self._hdr_dict is None:
            hdr_dict = OrderedDict()

            if self._pg is None:
                self._hdr_dict = hdr_dict
                return self._hdr_dict

            pg = self._pg
            hdr_dict['== pyFAI calibration =='] = ''
            hdr_dict['Spline file'] = pg.splineFile
            hdr_dict['Pixel size (m)'] = (pg.pixel1, pg.pixel2)
            hdr_dict['PONI (m)'] = (pg.poni1, pg.poni2)
            hdr_dict['Sample-detector distance (m)'] = pg.dist
            hdr_dict['Rotations (rad)'] = (pg.rot1, pg.rot2, pg.rot3)
            hdr_dict['wavelength'] = pg.wavelength

            hdr_dict['== pygix parameters =='] = ''
            hdr_dict['Sample orientation'] = pg.sample_orientation
            hdr_dict['Incident angle (degrees)'] = pg.incident_angle
            hdr_dict['Tilt angle (degrees)'] = pg.tilt_angle

            hdr_dict['== corrections =='] = ''
            if pg.maskfile is not None:
                hdr_dict['Mask file'] = pg.maskfile
            if has_dark or (pg.darkcurrent is not None):
                if pg.darkfiles:
                    hdr_dict['Dark current'] = pg.darkfiles
                else:
                    hdr_dict['Dark current'] = 'Unknown dark file'
            if has_flat or (pg.flatfield is not None):
                if pg.flatfiles:
                    hdr_dict['Flat field'] = pg.flatfiles
                else:
                    hdr_dict['Flat field'] = 'Unknown flat file'
            if (polarization_factor is None) and (pg._polarization is not None):
                polarization_factor = pg._polarization_factor
            hdr_dict['Polarization factor'] = polarization_factor
            hdr_dict['Normalization factor'] = normalization_factor
            self._hdr_dict = hdr_dict
        return self._hdr_dict

    def make_header_string(self, header_dict=None, hdr_key='#',
                           has_dark=False, has_flat=False,
                           polarization_factor=None, normalization_factor=None):
        """

        Args:
            header_dict:
            hdr_key:
            has_dark:
            has_flat:
            polarization_factor:
            normalization_factor:

        Returns:

        """
        if (header_dict is None) and (self._hdr_dict is None):
            self.make_header_dict(has_dark, has_flat, polarization_factor,
                                  normalization_factor)
        if header_dict is None:
            header_dict = self._hdr_dict

        hdr_list = []
        for k, v in header_dict.items():
            if (v.__class__ is str) and (len(v) is 0):
                hdr_list.append(k)
            else:
                hdr_list.append('{}: {}'.format(k, v))
        return '\n'.join(['{} {}'.format(hdr_key, i) for i in hdr_list])

    def make_header(self, hdr_key="#", has_dark=False, has_flat=False,
                    polarization_factor=None, normalization_factor=None):
        """

        Args:
            hdr_key (string): string used as comment in the header.
            has_dark (bool): save the file names of dark images.
            has_flat (bool): save the file names of flat images.
            polarization_factor (float):
            normalization_factor (float):

        Returns:

        """
        if self._header is None:
            if self._hdr_dict is None:
                self.make_header_dict(has_dark, has_flat, polarization_factor,
                                      normalization_factor)
                self._header = self.make_header_string(hdr_key=hdr_key)
        return self._header

    def save1D(self, filename, x_scale, intensity, error=None,
               x_unit=grazing_units.Q, has_dark=False, has_flat=False,
               polarization_factor=None, normalization_factor=None):
        """
        Save the data for 1D integration.

        Args:
            filename (string):
            x_scale (ndarray):
            intensity (ndarray):
            error (ndarray or None):
            x_unit (pygix.unit.Unit):
            has_dark (bool):
            has_flat (bool):
            polarization_factor (float or None):
            normalization_factor (float or None):
        """
        # x_unit = grazing_units.to_unit(x_unit)
        with open(filename, "w") as f:
            f.write(self.make_header(
                has_dark=has_dark, has_flat=has_flat,
                polarization_factor=polarization_factor,
                normalization_factor=normalization_factor))
            try:
                f.write("\n# --> %s\n" % filename)
            except UnicodeError:
                f.write("\n# --> %s\n" % (filename.encode("utf8")))
            if error is None:
                f.write("#%14s %14s\n" % (x_unit.REPR, "I "))
                f.write("\n".join(["%14.6e  %14.6e" % (t, i) for t, i in
                                   zip(x_scale, intensity)]))
            else:
                f.write("#%14s  %14s  %14s\n" %
                        (x_unit.REPR, "I ", "sigma "))
                f.write("\n".join(
                    ["%14.6e  %14.6e %14.6e" % (t, i, s) for t, i, s in
                     zip(x_scale, intensity, error)]))
            f.write("\n")

    def save2D(self, filename, intensity, x_scale=None, y_scale=None,
               error=None, hdr_note=None, has_dark=False, has_flat=False,
               polarization_factor=None, normalization_factor=None):
        """

        Args:
            filename:
            intensity:
            x_scale:
            y_scale:
            error:
            hdr_note (string):
            has_dark:
            has_flat:
            polarization_factor:
            normalization_factor:

        Returns:

        """
        if self._hdr_dict is None:
            self.make_header_dict(has_dark, has_flat, polarization_factor,
                                  normalization_factor)
        header = self._hdr_dict
        if hdr_note is not None:
            header['Note'] = hdr_note
        if (x_scale is not None) and (y_scale is not None):
            header['== data scaling =='] = ''
            header['x min'] = x_scale[0]
            header['x max'] = x_scale[-1]
            header['y min'] = y_scale[0]
            header['y max'] = y_scale[-1]

        try:
            img = fabio.edfimage.edfimage(data=intensity.astype("float32"),
                                          header=header)
            if error is not None:
                img.appendFrame(data=error,
                                header={"EDF_DataBlockID": "1.Image.Error"})
            img.write(filename)
        except IOError:
            print("IOError while writing %s" % filename)


def make_roi_header(**param):
    """
    Format header data to be written when saving ROI data.

    Args:
        method (string): integration method.
        param (dict): integration parameters.

    Returns:
        hdr_list (string): header data.
    """
    hdr_list = ['== Integration ROI ==']
    method = [i for i in param.keys() if "pos" in i][0].split('_pos')[0]
    hdr_list.append('Integration method: {}'.format(method))

    for k, v in param.items():
        hdr_list.append('{}: {}'.format(k, v))

    header = "\n".join(['# ' + i for i in hdr_list])
    return header


def save_roi(x_data, y_data, filename, **param):
    """


    Args:
        x_data (ndarray): x array of ROI.
        y_data (ndarray): y array of ROI.
        filename (string): path of file to be saved.
        **param (dict): integration parameters to be written the header.

    Returns:

    """
    header = make_roi_header(**param)

    with open(filename, 'w') as f:
        f.write(header)
        f.write('\n#%14s %14s\n' % ('x', 'y'))
        f.write('\n'.join(
            ['%14.6e  %14.6e' % (t, i) for t, i in zip(x_data, y_data)]))
        f.write('\n')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, division

__authors__ = ["Thomas Dane", "Jérôme Kieffer"]
__license__ = "GPLv3+"
__date__ = "18/03/2016"

import sys
import logging
import os

project = os.path.basename(os.path.dirname(os.path.abspath(__file__)))

try:
    from ._version import __date__ as date  # noqa
    from ._version import version, version_info, hexversion, strictversion  # noqa
except ImportError:
    raise RuntimeError("Do NOT use %s from its sources: build it and use the built version" % project)

logging.basicConfig()

if sys.version_info < (2, 6):
    logger = logging.getLogger("pygix.__init__")
    logger.error("pygix requires a python version >= 2.6")
    raise RuntimeError(
        "pygix requires a python version >= 2.6, now we are running: %s" %
        sys.version
        )
else:
    from .transform import Transform



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" A collection of useful tools for GIXS and fibre diffraction.

There are two main tools:

1. Four quadrant averaging:
    Fibre diffraction patterns are related through four quadrant symmetry
    (assuming ideal fibre texture). This means that the data in each of the
    four quadrants are equivalent.[1] This allows us to cut out each quadrant,
    centre, rotate and overlay, then average the sum of all four. This has two
    advantages. Firstly, statistics are improved. Secondly, regions of missing
    data (detector module gaps, asymmetric splitting due to large incident
    angle) are filled from other quadrants - prettier images!

    Example usage:
        averaged, x, y = quadrant_average(data, x, y)

    N.B. The span of the image will change and will be determined by
    -absmax(q)) to +absmax(q) such that new x and y scales are returned.

2. Integration ROIs:
    Arrays defining the regions-of-interest (ROIs) for 1D data reduction methods
    can be calculated. These are useful for overlaying on diffraction patterns
    in figures to show the regions of data that have been integrated. Four
    methods exist for each of the four 1D reduction methods:
        sector_roi
        chi_roi
        op_box_roi
        ip_box_roi

    These functions take the same parameters as the integration methods.

    N.B. These are all calculated in the output coordinate space, so must be
    overlayed onto transformed images, NOT raw images.

References:
    [1] Stribeck and Nöchel, J. Appl. Crystallogr., (2009), 42, 295.
"""

import numpy as np
from . import io


def quadrant_average(data, x=None, y=None, dummy=0, filename=None):
    """
    Function to perform four quadrant averaging of fiber diffraction
    patterns. If only the data array is given, the function 
    assumes the center of reciprocal space is the central pixel
    and the returned averaged array will have the same shape as 
    the input data. In general this should be avoided, it is safest if
    x and y scales are provided!

    If x and y scaling are given, it deduces the
    center from these arrays. One quadrant will have the size of 
    the largest quadrant and the resulting array will be larger 
    than the input array. In this case a new x and y will be
    calculated and returned.
    
    Args:
        data (ndarray): Input reciprocal space map array.
        x (ndarray): x scaling of the image.
        y (ndarray: y scaling of the image.
        dummy (int): Value of masked invalid regions.
        filename (string): Name of file to be saved.

    Returns:
        out_full (ndarray): The four quadrant averaged array
    """
    if (x is not None) and (y is not None):
        cen_x = np.argmin(abs(x))
        cen_y = np.argmin(abs(y))
    elif (x is None) and (y is None):
        cen_y = data.shape[0] / 2.0
        cen_x = data.shape[1] / 2.0
    else:
        raise RuntimeError('Must pass both x and y scales or neither')

    quad1 = np.flipud(np.fliplr(data[0:cen_y, 0:cen_x]))
    quad2 = np.fliplr(data[cen_y:, 0:cen_x])
    quad3 = data[cen_y:, cen_x:]
    quad4 = np.flipud(data[0:cen_y, cen_x:])

    quad_shape_y = max(data.shape[0] - cen_y,
                       data.shape[0] - (data.shape[0] - cen_y))
    quad_shape_x = max(data.shape[1] - cen_x,
                       data.shape[1] - (data.shape[1] - cen_x))
    mask = np.zeros((quad_shape_y, quad_shape_x))
    out = np.zeros((quad_shape_y, quad_shape_x))

    out[np.where(quad1 > dummy)] += quad1[np.where(quad1 > dummy)]
    out[np.where(quad2 > dummy)] += quad2[np.where(quad2 > dummy)]
    out[np.where(quad3 > dummy)] += quad3[np.where(quad3 > dummy)]
    out[np.where(quad4 > dummy)] += quad4[np.where(quad4 > dummy)]

    mask[np.where(quad1 > dummy)] += 1
    mask[np.where(quad2 > dummy)] += 1
    mask[np.where(quad3 > dummy)] += 1
    mask[np.where(quad4 > dummy)] += 1

    out[np.where(mask > 0)] /= mask[np.where(mask > 0)]
    out[np.where(mask == 0)] = dummy

    out_full = np.zeros((out.shape[0] * 2, out.shape[1] * 2))
    cen_x = out_full.shape[1] / 2.0
    cen_y = out_full.shape[0] / 2.0

    out_full[0:cen_y, 0:cen_x] = np.flipud(np.fliplr(out))
    out_full[0:cen_y, cen_x:] = np.flipud(out)
    out_full[cen_y:, cen_x:] = out
    out_full[cen_y:, 0:cen_x] = np.fliplr(out)

    if (x is not None) and (y is not None):
        out_x = np.linspace(-abs(x).max(), abs(x).max(), out_full.shape[1])
        out_y = np.linspace(-abs(y).max(), abs(y).max(), out_full.shape[0])
    else:
        out_x = None
        out_y = None

    if filename is not None:
        writer = io.Writer(None, None)
        writer.save2D(filename, out_full, out_x, out_y)

    if (out_x is not None) and (out_y is not None):
        return out_full, out_x, out_y
    else:
        return out_full


def sector_roi(chi_pos=None, chi_width=None, radial_range=None, filename=None):
    """Generate array defining region of interest for sector integration.

    Args:
        chi_pos (float): chi angle (deg) defining the centre of the sector.
        chi_width (float): width (deg) of sector.
        radial_range (tuple): integration range (min, max).
        filename (string): filename to save the arrays.

    Returns:
        qr, qz (tuple of ndarrays): arrays defining the region of interest.
    """
    param = locals()  # passed only to io to write key, val in the header
    param.pop('filename')

    if (len([x for x in [chi_pos, chi_width, radial_range] if
             x is not None]) is 0) \
            or (radial_range is None):
        raise RuntimeError('Integration over whole image, no ROI to display.')
    roi_x, roi_y = _calc_sector(radial_range, chi_pos, chi_width)

    if filename:
        io.save_roi(roi_x, roi_y, filename, **param)
    return roi_x, roi_y


def chi_roi(radial_pos, radial_width, chi_range=None, filename=None):
    """Generate array defining region of interest for chi integration.

    Args:
        radial_pos (float): position defining the radius of the sector.
        radial_width (float): width (q or 2th) of sector.
        chi_range (tuple): azimuthal range (min, max).
        filename (string): filename to save the arrays.

    Returns:
        qr, qz (tuple of ndarrays): arrays defining the region of interest.
    """
    param = locals()  # passed only to io to write key, val in the header
    param.pop('filename')

    if (chi_range is None) or (chi_range[0] + chi_range[1] is 360):
        chi_width = None
        chi_pos = None
    else:
        chi_width = chi_range[1] - chi_range[0]
        chi_pos = chi_range[0] + chi_width / 2.0

    radial_min = radial_pos - radial_width / 2.0
    radial_max = radial_pos + radial_width / 2.0
    roi_x, roi_y = _calc_sector((radial_min, radial_max), chi_pos, chi_width)

    if filename:
        io.save_roi(roi_x, roi_y, filename, **param)
    return roi_x, roi_y


def op_box_roi(ip_pos, ip_width, op_range, filename=None):
    """Generate array defining region of interest for out-of-plane box integration.

    Args:
        ip_pos (float): in-plane centre of integration box.
        ip_width (float): in-plane width of integration box.
        op_range (tuple): out-of-plane range (min, max).
        filename (string): filename to save the arrays.

    Returns:
        qr, qz (tuple of ndarrays): arrays defining the region of interest.
    """
    param = locals()  # passed only to io to write key, val in the header
    param.pop('filename')

    ip_min = ip_pos - ip_width / 2.0
    ip_max = ip_pos + ip_width / 2.0
    roi_x, roi_y = _calc_box((ip_min, ip_max), op_range)

    if filename:
        io.save_roi(roi_x, roi_y, filename, **param)
    return roi_x, roi_y


def ip_box_roi(op_pos, op_width, ip_range, filename=None):
    """Generate array defining region of interest for in-plane box integration.

    Args:
        op_pos (float): out-of-plane centre of integration box.
        op_width (float): out-of-plane width of integration box.
        ip_range (tuple): in-plane range (min, max).
        filename (string): filename to save the arrays.

    Returns:
        qr, qz (tuple of ndarrays): arrays defining the region of interest.
    """
    param = locals()  # passed only to io to write key, val in the header
    param.pop('filename')

    op_min = op_pos - op_width / 2.0
    op_max = op_pos + op_width / 2.0
    roi_x, roi_y = _calc_box(ip_range, (op_min, op_max))

    if filename:
        io.save_roi(roi_x, roi_y, filename, **param)
    return roi_x, roi_y


def _calc_sector(radial_range, chi_pos, chi_width):
    """Main function for calculating sector region of interest.
    Called by sector_roi and chi_roi.

    Args:
        radial_range (tuple): integration range (min, max).
        chi_pos (float): chi angle (deg) defining the centre of the sector.
        chi_width (float): width (deg) of sector.

    Returns:
        qr, qz (tuple of ndarrays): arrays defining the region of interest.
    """
    if len([x for x in [chi_pos, chi_width] if x is not None]) is 1:
        raise RuntimeError('both chi_pos and chi_width must be supplied or '
                           'neither')

    if (chi_pos is None) and (chi_width is None):
        chi_min = 0
        chi_max = 359
        npts = 360
    else:
        chi_min = -(chi_pos - chi_width / 2.0 - 90.0)
        chi_max = -(chi_pos + chi_width / 2.0 - 90.0)
        npts = abs(int(chi_max - chi_min))

    chi = np.radians(np.linspace(chi_min, chi_max, npts))

    # lower part of arc
    if radial_range[0] is 0:
        lo_qr = np.array(0)
        lo_qz = np.array(0)
    else:
        lo_qr = radial_range[0] * np.cos(chi)
        lo_qz = radial_range[0] * np.sin(chi)

    # upper part of arc
    hi_qr = (radial_range[1] * np.cos(chi))[::-1]
    hi_qz = (radial_range[1] * np.sin(chi))[::-1]

    qr = np.hstack((lo_qr, hi_qr))
    qz = np.hstack((lo_qz, hi_qz))

    if (chi_pos is not None) and (chi_width is not None):
        qr = np.append(qr, qr[0])
        qz = np.append(qz, qz[0])
    return qr, qz


def _calc_box(ip_range, op_range):
    """Main function for calculating box regions of interest.
    Called by op_box_roi and ip_box_roi.

    Args:
         ip_range (tuple): in-plane (min, max).
        op_range (tuple): out-of-plane (min, max).

    Returns:
        qr, qz (tuple of ndarrays): arrays defining the region of interest.
    """
    qr = [ip_range[0], ip_range[0], ip_range[1], ip_range[1], ip_range[0]]
    qz = [op_range[0], op_range[1], op_range[1], op_range[0], op_range[0]]
    return qr, qz



#!/usr/bin/python
# coding: utf8
# /*##########################################################################
#
# Copyright (c) 2015-2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["Jérôme Kieffer", "Thomas Vincent"]
__date__ = "18/03/2016"
__license__ = "MIT"


import sys
import os

from numpy.distutils.misc_util import Configuration

try:
    from setuptools import setup
    from setuptools.command.build_py import build_py as _build_py
    from setuptools.command.build_ext import build_ext
    from setuptools.command.sdist import sdist
except ImportError:
    from numpy.distutils.core import setup
    from distutils.command.build_py import build_py as _build_py
    from distutils.command.build_ext import build_ext
    from distutils.command.sdist import sdist

PROJECT = "pygix"
cmdclass = {}


# Check if action requires build/install
DRY_RUN = len(sys.argv) == 1 or (len(sys.argv) >= 2 and (
    '--help' in sys.argv[1:] or
    sys.argv[1] in ('--help-commands', 'egg_info', '--version',
                    'clean', '--name')))


def get_version():
    import version
    return version.strictversion


def get_readme():
    dirname = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(dirname, "README.md"), "r", encoding='utf-8') as fp:
        long_description = fp.read()
    return long_description


classifiers = ["Development Status :: 2 - Pre-Alpha",
               "Intended Audience :: Developers",
               "Intended Audience :: Education",
               "Intended Audience :: Science/Research",
               "Natural Language :: English",
               "Programming Language :: Cython",
               "Programming Language :: Python :: 3",
               "Programming Language :: Python :: Implementation :: CPython",
               "Environment :: Console",
               "Environment :: MacOS X",
               "Environment :: Win32 (MS Windows)",
               "License :: OSI Approved :: MIT License",
               "Operating System :: Microsoft :: Windows",
               "Operating System :: Unix",
               "Operating System :: POSIX",
               "Operating System :: MacOS :: MacOS X",
               "Topic :: Documentation :: Sphinx",
               "Topic :: Scientific/Engineering :: Physics",
               "Topic :: Software Development :: Libraries :: Python Modules",
               ]


# ########## #
# version.py #
# ########## #


class build_py(_build_py):
    """
    Enhanced build_py which copies version.py to <PROJECT>._version.py
    """
    def find_package_modules(self, package, package_dir):
        modules = _build_py.find_package_modules(self, package, package_dir)
        if package == PROJECT:
            modules.append((PROJECT, '_version', 'version.py'))
        return modules


cmdclass['build_py'] = build_py


# ################### #
# build_doc commands  #
# ################### #

try:
    import sphinx
    import sphinx.util.console
    sphinx.util.console.color_terminal = lambda: False
    from sphinx.setup_command import BuildDoc
except ImportError:
    sphinx = None
else:
    # i.e. if sphinx:
    class build_doc(BuildDoc):

        def run(self):
            # make sure the python path is pointing to the newly built
            # code so that the documentation is built on this and not a
            # previously installed version

            build = self.get_finalized_command('build')
            sys.path.insert(0, os.path.abspath(build.build_lib))

#             # Copy .ui files to the path:
#             dst = os.path.join(
#                 os.path.abspath(build.build_lib), "silx", "gui")
#             if not os.path.isdir(dst):
#                 os.makedirs(dst)
#             for i in os.listdir("gui"):
#                 if i.endswith(".ui"):
#                     src = os.path.join("gui", i)
#                     idst = os.path.join(dst, i)
#                     if not os.path.exists(idst):
#                         shutil.copy(src, idst)

            # Build the Users Guide in HTML and TeX format
            for builder in ('html', 'latex'):
                self.builder = builder
                self.builder_target_dir = os.path.join(self.build_dir, builder)
                self.mkpath(self.builder_target_dir)
                BuildDoc.run(self)
            sys.path.pop(0)
    cmdclass['build_doc'] = build_doc


# ############################# #
# numpy.distutils Configuration #
# ############################# #


def configuration(parent_package='', top_path=None):
    """Recursive construction of package info to be used in setup().
    See http://docs.scipy.org/doc/numpy/reference/distutils.html#numpy.distutils.misc_util.Configuration
    """  # noqa
    config = Configuration(None, parent_package, top_path)
    config.set_options(
        ignore_setup_xxx_py=True,
        assume_default_configuration=True,
        delegate_options_to_subpackages=True,
        quiet=True)
    config.add_subpackage(PROJECT)
    return config


config = configuration()


################################################################################
# Debian source tree
################################################################################


class sdist_debian(sdist):
    """
    Tailor made sdist for debian
    * remove auto-generated doc
    * remove cython generated .c files
    """
    def prune_file_list(self):
        sdist.prune_file_list(self)
        to_remove = ["doc/build", "doc/pdf", "doc/html", "pylint", "epydoc"]
        print("Removing files for debian")
        for rm in to_remove:
            self.filelist.exclude_pattern(pattern="*", anchor=False, prefix=rm)

        # this is for Cython files specifically: remove C & html files
        search_root = os.path.dirname(os.path.abspath(__file__))
        for root, _, files in os.walk(search_root):
            for afile in files:
                if os.path.splitext(afile)[1].lower() == ".pyx":
                    base_file = os.path.join(root, afile)[len(search_root) + 1:-4]
                    self.filelist.exclude_pattern(pattern=base_file + ".c")
                    self.filelist.exclude_pattern(pattern=base_file + ".cpp")
                    self.filelist.exclude_pattern(pattern=base_file + ".html")

    def make_distribution(self):
        self.prune_file_list()
        sdist.make_distribution(self)
        dest = self.archive_files[0]
        dirname, basename = os.path.split(dest)
        base, ext = os.path.splitext(basename)
        while ext in [".zip", ".tar", ".bz2", ".gz", ".Z", ".lz", ".orig"]:
            base, ext = os.path.splitext(base)
        if ext:
            dest = "".join((base, ext))
        else:
            dest = base
        sp = dest.split("-")
        base = sp[:-1]
        nr = sp[-1]
        debian_arch = os.path.join(dirname, "-".join(base) + "_" + nr + ".orig.tar.gz")
        os.rename(self.archive_files[0], debian_arch)
        self.archive_files = [debian_arch]
        print("Building debian .orig.tar.gz in %s" % self.archive_files[0])

cmdclass['debian_src'] = sdist_debian


# ##### #
# setup #
# ##### #

setup_kwargs = config.todict()


install_requires = ["numpy", "pyFAI"]
setup_requires = ["numpy"]

setup_kwargs.update(
    name=PROJECT,
    version="2021.10.7.1",
    url="https://github.com/tgdane/pygix",
    author="Thomas Dane",
    author_email="thomasgdane@gmail.com",
    classifiers=classifiers,
    description="Software library for grazing-incidence/fibre X-Ray data analysis",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    install_requires=install_requires,
    setup_requires=setup_requires,
    cmdclass=cmdclass,
    )

setup(**setup_kwargs)
